//	BLLoad.hh

#ifndef BLLOAD_HH
#define BLLOAD_HH


#include "globals.hh"

// get the proper extension for the current OS
#ifdef __linux
#define DLLOAD_EXT ".so"
#include <dlfcn.h>      // for dlopen()
#endif
#ifdef WIN32
#define DLLOAD_EXT ".dll"
#endif
#ifdef __CYGWIN__
#define DLLOAD_EXT ".dll"
#endif
#ifdef __APPLE__
#define DLLOAD_EXT ".dylib"
#endif
#ifndef DLLOAD_EXT
#error unsupported OS
#endif

/** class BLLoad will load a shared object.
 *
 *  All static object constructors in the shared object are executed
 *  while loading, and they can register the resources implemented.
 *
 *  Optional features are implemented in shared objects that are
 *  automatically loaded when used. User-implemented shared objects
 *  must be explicitly loaded (see BLCMDload).
 *
 *  All functions are inline, so no .cc file is needed.
 **/
class BLLoad {
public:
	/// load() loads a shared object.
	/// If no '/' or '\' is present in the name, it will be
	/// searched for in the usual way.
	/// Normally no extension is used (.so, .dylib, or .dll);
	/// this function then appends the proper extension for the 
	/// current OS. Loading an already loaded shared object 
	/// does nothing (no error). If file is not found, tries
	/// prepending "lib" to the name.
	/// Returns true on success, false on error.
	static bool load(G4String name) {
#ifdef WIN32
		return false;
#endif
#ifdef __APPLE__
		return false;
#endif
#ifdef __CYGWIN__
		return false;
#endif
#ifdef __linux
		if(name.find(DLLOAD_EXT) == G4String::npos)
			name.append(DLLOAD_EXT);
		if(!dlopen(name.c_str(),RTLD_NOW)) {
			name = G4String("lib")+name;
			if(!dlopen(name.c_str(),RTLD_NOW))
				return false;
		}
		return true;
#endif // ! WIN32
	}

	/// errorString() returns a string describing the previous 
	/// error.
	static const char *errorString() {
#ifdef WIN32
		return "Unsupported OS";
#endif
#ifdef __APPLE__
		return "Unsupported OS";
#endif
#ifdef __CYGWIN__
		return "Unsupported OS";
#endif
#ifdef __linux
		return dlerror(); 
#endif
	}
};

#endif // BLLOAD_HH
