//	BLMinimize.hh
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/


#ifndef BLMINIMIZE_HH
#define BLMINIMIZE_HH

#include <stdio.h>
#include <vector>

#include <gsl/gsl_multimin.h>

/**	class BLMinimizeFunction is the base class for a class defining a
 *	function to minimize.
 **/
class BLMinimizeFunction {
public:
	BLMinimizeFunction() { }
	virtual ~BLMinimizeFunction() { }

	/// operator() is the function to minimize. Note it is not const
	/// for flexibility in derived classes.
	/// Its argument is necessarily const.
	virtual double operator()(const std::vector<double> &x) = 0;
};

/**	class BLMinimize implements a general function minimizer using
 *	the GSL multimin functions.
 *
 *	Uses the the Simplex algorithm of Nelder and Mead.
 *
 *	Compilation for simple programs:
 *		g++ -I... test.cc -lgsl -lgslcblas -lm
 *
 *	NOTE: the parameters to the minimizer must all have scales that are
 *	comparable, or the minimizer will often converge to a wildly incorrect
 *	value. This applies to track fitting, where positions have scales of
 *	1 (millimeters), but angles have scales of 0.001 (milliradians).
 *	This class scales the parameters appropriately. it also permits the
 *	function to have arguments that are held constant during the 
 *	minimization by setting their scale to 0.0.
 *
 *	This class is entirely inline.
 **/
class BLMinimize {
	std::vector<double> scale;
	std::vector<double> arg;
	double val;
	double size;
	int iter;
	int print;
	double chisqMin;
	BLMinimizeFunction *func;
	gsl_multimin_function ff;
	gsl_multimin_fminimizer *m;
private:
	// f() is the function the gsl routines call (they are in C)
	static double f(const gsl_vector *x, void *param) {
		BLMinimize *p = (BLMinimize*)param;
		return p->call(x);
	}
	// call() will call the user's function, scaling the argument vector.
	double call(const gsl_vector *x) {
		if(print >= 2) printf("BLMinimize call args:");
		int j=0;
		for(unsigned i=0; i<scale.size(); ++i) {
			if(scale[i] != 0.0) // fixed args are already set
				arg[i] = gsl_vector_get(x,j++) * scale[i];
			if(print >= 2)
				printf(" %.4g",arg[i]);
		}
		if(print >= 2) printf("\n");
		double v = (*func)(arg);
		if(print >= 2) printf("BLMinimize call returns value=%.4g\n",v);
		return v;
	}
public:
	/// Constructor.
	BLMinimize(int _print=2) : scale(), arg()
		{ val=GSL_NAN; size=GSL_NAN; iter=0; print=_print;
		  chisqMin=GSL_NAN; func=0; ff.f=0; ff.n=0; ff.params=0;  m=0;
		}

	/// Constructor. Sets function to minimize and the scale of its
	/// arguments, and optionally print.
	BLMinimize(BLMinimizeFunction &f, const std::vector<double> &_scale,
						int _print=2) : scale(), arg()
		{ val=GSL_NAN; size=GSL_NAN; iter=0; print=_print;
		  chisqMin=GSL_NAN; func=0; ff.f=0; ff.n=0; ff.params=0;  m=0;
		  setFunc(f,_scale);
		}

	virtual ~BLMinimize() { if(m) gsl_multimin_fminimizer_free(m); }

	/// setFunc() will set the function to minimize and the scale of its
	/// arguments. The number of non-zero entries in _scale[] determines
	/// the number of variables to minimize.
	void setFunc(BLMinimizeFunction &f, const std::vector<double> &_scale) {
		func = &f;
		scale.clear();
		arg.clear();
		int nx = 0;
		for(unsigned i=0; i<_scale.size(); ++i) {
			double s=_scale[i];
			scale.push_back(s);
			arg.push_back(s);
			if(s != 0.0) ++nx;
		}
		ff.f=BLMinimize::f;
		ff.n=nx;
		ff.params=this;
		if(m) gsl_multimin_fminimizer_free(m);
		m = 0;
		if(nx == 0) return;
		m = gsl_multimin_fminimizer_alloc(
		  	gsl_multimin_fminimizer_nmsimplex,nx);
	}

	/// setPrintDetail() sets the detail of printing per iteration:
	/// 0=no print, 1=print value and size only, 2=print value, size, and x.
	/// the default is 2.
	void setPrintDetail(int v=0) { print = v; }

	/// setChisqMin() sets an additional convergence criterium of chisq
	/// less than the given value. Can only be used when the function is
	/// positive definite, usually a chi-squared. Required for track-
	/// fitting of test tracks that have truly 0 chisq -- the normal 
	/// convergence often fails for such tracks.
	void setChisqMin(double v) { chisqMin = v; }

	/// minimize() finds the minimum, returning status (0=success).
	///   x		the initial and then the final argument values
	///   tolerance	the convergence limit on the simplex size
	///		NOTE: this is after scaling!
	///   maxIter	limits the number of iterations
	/// The initial step for x[i] is scale[i] (x[i] is scaled, so the
	/// step to the gsl routine is 1.0).
	int minimize(std::vector<double> &x, double tolerance=1e-3,
							int maxIter=100) {
		if(ff.n <= 0) return GSL_EINVAL;
		assert(x.size()==scale.size());
		gsl_vector *init = gsl_vector_alloc(ff.n);
		gsl_vector *istep = gsl_vector_alloc(ff.n);
		unsigned j=0;
		for(unsigned i=0; i<scale.size(); ++i) {
			arg[i] = x[i]; // fixed args, variable ones overwritten
			if(scale[i] != 0.0) {
				gsl_vector_set(init,j,x[i]/scale[i]); 
				gsl_vector_set(istep,j,1.0);
				++j;
			}
		}
		assert(j==ff.n); assert(m!=0);
		gsl_multimin_fminimizer_set(m,&ff,init,istep);
		for(iter=1; iter<=maxIter; ++iter) {
			int status = gsl_multimin_fminimizer_iterate(m);
			if(status) {
				if(print >= 1)
					printf("BLMinimize FAILED status=%d\n",
									status);
				gsl_vector_free(init);
				gsl_vector_free(istep);
				return status;
			}
			val = m->fval;
			size = gsl_multimin_fminimizer_size(m);
			if(print >= 1)
			    printf("BLMinimize Iter %d: val=%.4g size=%.4g\n",
			    				iter,val,size);
			status = gsl_multimin_test_size(size,tolerance);
			if(status == GSL_SUCCESS || (!gsl_isnan(chisqMin) &&
			   val <= chisqMin)) {
				unsigned j=0;
				for(unsigned i=0; i<scale.size(); ++i) {
				    if(scale[i] != 0.0)
					x[i] = gsl_vector_get(m->x,j++) *
								scale[i];
				}
				gsl_vector_free(init);
				gsl_vector_free(istep);
				if(print >= 1) printf("BLMinimize SUCCESS\n");
				return 0;
			}
		}
		gsl_vector_free(init);
		gsl_vector_free(istep);
		if(print >= 1) printf("BLMinimize FAILED iter > %d\n",maxIter);
		return GSL_EMAXITER;
	}

	/// value() returns the function value from the previous minimize().
	double value() const { return val; }

	/// getSize() returns the Simplex size from the previous minimize().
	double getSize() { return size; }

	/// getIterations() returns the # iterations of the previous minimize().
	int getIterations() { return iter; }
};

#endif // BLMINIMIZE_HH
