//	BLCMDcollective.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>

#include "BLCommand.hh"
#include "BLCollectiveComputation.hh"
#include "BLRunManager.hh"
#include "BLNTuple.hh"
#include "BLParam.hh"

/**	class BLCMDcollective - example collective computation.
 *
 *	This class simply computes the mean and sigma of the time stepping
 *	of BLRunManager, and enters them into an NTuple.
 **/
class BLCMDcollective : public BLCommand, public BLCollectiveComputation {
	G4double deltaT;
	G4int verbose;
	BLNTuple *timeNtuple;
	BLRunManager *runManager;
public:
	BLCMDcollective();

	G4String commandName() { return "collective"; }
	
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	void defineNamedArgs();

	virtual void beginCollectiveTracking(std::vector<BLTrackData>& v);

	virtual void collectiveStep(std::vector<BLTrackData>& v);

	virtual void endCollectiveTracking(std::vector<BLTrackData>& v);
};

BLCMDcollective defaultCollectiveCommand;

BLCMDcollective::BLCMDcollective()
{
	registerCommand(BLCMDTYPE_OTHER);
	setSynopsis("Example collective computation");
	setDescription("This command computes the mean and sigma related "
		"to the time stepping in BLRunManager. This class is "
		"intended to serve as an example for a collective "
		"computation.");

	// initialize class variables here
	deltaT = 1.0*ns;
	verbose = -1;
	timeNtuple = 0;
	runManager = 0; // not available, yet -- see command(...)
}

int BLCMDcollective::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	runManager = BLRunManager::getObject();

	handleNamedArgs(namedArgs);

	if(verbose < 0) verbose=Param.getInt("steppingVerbose");

	runManager->registerCollectiveComputation(this);
	runManager->setCollectiveMode(true);
	runManager->setDeltaT(deltaT);

	print("");

	return 0;
}

void BLCMDcollective::defineNamedArgs()
{
	argDouble(deltaT,"deltaT","Time step (1 ns).",ns);
	argInt(verbose,"verbose","Non-zero for verbose prints (0).");
}	

void BLCMDcollective::beginCollectiveTracking(std::vector<BLTrackData>& v)
{
	if(verbose) printf("beginCollectiveTracking v.size=%d\n",
		(int)v.size());

	// initialize the computation here.
	timeNtuple = BLNTuple::create("", "Ntuples", "TimeStep",
				"T:Ntracks:meanDT:sigmaDT","TimeStep");
}

void BLCMDcollective::collectiveStep(std::vector<BLTrackData>& v)
{
	// init variables for the computation
	G4double stepTime = runManager->getNextStepTime();
	G4double sumDT=0.0, sumDT2=0.0;

	// loop over tracks
	int active=0;
	for(unsigned i=0; i<v.size(); ++i) {
		// get the track pointer, and verify it is alive
		G4Track *track = v[i].track;
		G4TrackStatus trackStatus = track->GetTrackStatus();
		if(trackStatus != fAlive && trackStatus != fStopButAlive)
			continue;
		++active;

		// get basic track data
		int evId = v[i].event->GetEventID();
		int trackId = BLManager::getObject()->
					getExternalTrackID(track->GetTrackID());
		G4ThreeVector position = track->GetPosition();
		G4ThreeVector momentum = track->GetMomentum();
		G4double time = track->GetGlobalTime();

		// do the computation for this track
		G4double dt = time - stepTime;
		sumDT += dt;
		sumDT2 += dt*dt; 
	}

	// complete the computation
	G4double mean=0.0, sigma=0.0;
	if(active >= 2) {
		mean = sumDT/active;
		sigma = sqrt(sumDT2/active - mean*mean);
		float field[4];
		field[0] = stepTime;
		field[1] = active;
		field[2] = mean;
		field[3] = sigma;
		if(timeNtuple) timeNtuple->appendRow(field,sizeof(field)/sizeof(field[0]));
	}

	if(verbose) {
		printf("CollectiveStep v.size=%d active=%d time=%.3f mean=%.3f sigma=%.3f\n",
			(int)v.size(),active,
			runManager->getNextStepTime(), 
			mean,sigma);
	}
}

void BLCMDcollective::endCollectiveTracking(std::vector<BLTrackData>& v)
{
	if(verbose) printf("endCollectiveTracking v.size=%d\n",(int)v.size());

	// finalize and summarize the computation here
}
