//	BLCMDmovie.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>

#include "BLCommand.hh"
#include "BLNTuple.hh"
#include "BLParam.hh"
#include "BLCoordinates.hh"

const char OutlineFields[] =
 "type:Xcl:Ycl:Zcl:Xg:Yg:Zg:Height:Width:Length:XZAngle:YZangle:Red:Green:Blue";
const int NOutlineFields = 15;

const char ReferenceFields[] = "T:Zref:Xcl:Ycl:Zcl:Xg:Yg:Zg:Ptot";
const int NReferenceFields = 9;

const char TrackFields[] =
    "x:y:z:Px:Py:Pz:t:PDGid:EventID:TrackID:ParentID:Weight";
const unsigned NTrackFields = 12;

/**	class BLCMDmovie - output NTuple-s suitable for a movie
 *
 **/
class BLCMDmovie : public BLCommand, public BLCallback, 
					public BLManager::SteppingAction,
					public BLManager::TrackingAction {
	BLNTuple *outline;	// NTuple giving outlines of beamline elements
	BLNTuple *reference;	// NTuple giving reference particle coords
	BLNTuple *trace;	// NTuple tracing all beam tracks
	BLManager *manager;
public:
	BLCMDmovie();

	G4String commandName() { return "movie"; }
	
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	void defineNamedArgs();

	/// callback() from BLCallback.
	void callback(int type);

	/// UserSteppingAction() from BLManager::SteppingAction.
	void UserSteppingAction(const G4Step *step);

	/// from BLManager::TrackingAction
	virtual void PreUserTrackingAction(const G4Track *track);
	virtual void PostUserTrackingAction(const G4Track *track);

	void referenceStep(const G4Track *track);
	void beamStep(const G4Track *track);
};

BLCMDmovie defaultMovieCommand;

BLCMDmovie::BLCMDmovie() : BLCallback(), BLManager::SteppingAction()
{
	registerCommand(BLCMDTYPE_OTHER);
	setSynopsis("Generate movie NTuple.");
	setDescription("This command outputs a set of NTuples suitable for "
		"generating a movie.");

	// initialize class variables here
	outline = 0;
	reference = 0;
	trace = 0;
	manager = 0;
}

int BLCMDmovie::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	manager = BLManager::getObject();

	handleNamedArgs(namedArgs);

	print("");

	// register reference coords before registering ourselves
	BLCoordinates::useReferenceCoordinates();

	manager->registerTrackingAction(this);
	manager->registerReferenceParticleStep(0,this);
	manager->registerBeamStep(0,this);
	manager->registerCallback(this,0);
	manager->registerCallback(this,2);

	//outline = BLNTuple::create("root", "Movie", "Outline",
	//						OutlineFields, "");
	reference = BLNTuple::create("root", "Movie", "Reference", 
							ReferenceFields, "");
	trace = BLNTuple::create("root", "Movie", "Tracks", 
							TrackFields, "");

	return 0;
}

void BLCMDmovie::defineNamedArgs()
{
}	

void BLCMDmovie::callback(int type)
{
	if(type == 0)  {	// pre-tune callback
	} else if(type == 2) {	// post-beam callback
	}
}

void BLCMDmovie::UserSteppingAction(const G4Step *step)
{
	switch(manager->getState()) {
	case IDLE:
		break;
	case VISUAL:
		break;
	case TUNE:
		break;
	case REFERENCE:
		referenceStep(step->GetTrack());
		break;
	case BEAM:
		beamStep(step->GetTrack());
		break;
	case SPECIAL:
		break;
	}
}

void BLCMDmovie::PreUserTrackingAction(const G4Track *track)
{
	if(manager->getState() == REFERENCE) referenceStep(track);
	//@ if(manager->getState() == BEAM) beamStep(track);
}

void BLCMDmovie::PostUserTrackingAction(const G4Track *track)
{
	if(manager->getState() == REFERENCE) referenceStep(track);
	if(manager->getState() == BEAM) beamStep(track);
}

void BLCMDmovie::referenceStep(const G4Track *track)
{
	G4ThreeVector posGlobal = track->GetPosition();
	G4ThreeVector posCL = posGlobal;
	G4double Zref = BLCoordinates::getMostRecentReferenceZ();
	G4double time = track->GetGlobalTime();
	G4ThreeVector momentum = track->GetMomentum();

	// transform to centerline coordinates
	BLCoordinates *c = (BLCoordinates *)track->GetUserInformation();
	if(c && c->isValid()) {
		c->getCoords(BLCOORD_CENTERLINE,posCL);
	}

	assert(NReferenceFields == 9);
	float data[NReferenceFields];
	data[0] = time/ns;
	data[1] = Zref/mm;
	data[2] = posCL[0]/mm;
	data[3] = posCL[1]/mm;
	data[4] = posCL[2]/mm;
	data[5] = posGlobal[0]/mm;
	data[6] = posGlobal[1]/mm;
	data[7] = posGlobal[2]/mm;
	data[8] = momentum.mag()/MeV;

	reference->appendRow(data,NReferenceFields);
}

void BLCMDmovie::beamStep(const G4Track *track)
{
	G4RunManager* runmgr = G4RunManager::GetRunManager();
	const G4Event* event = runmgr->GetCurrentEvent();
	int evNum = event->GetEventID();
	G4ThreeVector position = track->GetPosition();
	G4double time = track->GetGlobalTime();
	G4ThreeVector momentum = track->GetMomentum();

	// transform to reference coordinates
	BLCoordinates *c = (BLCoordinates *)track->GetUserInformation();
	if(c && c->isValid()) {
		c->getCoords(BLCOORD_REFERENCE,position);
		momentum = c->getRotation() * momentum;
	}

	assert(NTrackFields == 12);
	float data[NTrackFields];
	data[0] = position[0]/mm;		// x (mm)
	data[1] = position[1]/mm;		// y (mm)
	data[2] = position[2]/mm;		// z (mm)
	data[3] = momentum[0]/MeV;		// Px (MeV/c)
	data[4] = momentum[1]/MeV;		// Py (MeV/c)
	data[5] = momentum[2]/MeV;		// Pz (MeV/c)
	data[6] = time/ns;			// t (ns)
	data[7] = track->GetDefinition()->GetPDGEncoding();
	data[8] = evNum;				// Event ID
	data[9] = 
		BLManager::getObject()->getExternalTrackID(track->GetTrackID());
	data[10] = track->GetParentID();	// Parent ID
	data[11] = track->GetWeight();		// Weight

	trace->appendRow(data,NTrackFields);
}

