//	BLCMDnewparticlentuple.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include "BLNTuple.hh"
#include "BLManager.hh"
#include "BLCommand.hh"
#include "BLCoordinates.hh"
#include "BLEvaluator.hh"

const char TrackFields[] =
    "x:y:z:Px:Py:Pz:t:PDGid:EventID:TrackID:ParentID:Weight";
const unsigned NTrackFields = 12;

/**	class BLCMDnewparticlentuple - NTuple for tracks where they are created.
 *	
 **/
class BLCMDnewparticlentuple : public BLCommand,
					public BLManager::TrackingAction {
	G4String name;
	G4String format;
	G4String filename;
	G4String require;
	G4String coordinates;
	BLCoordinateType coordinateType;
	BLNTuple *ntuple;
	BLEvaluator *eval;
public:
	BLCMDnewparticlentuple();

	BLCMDnewparticlentuple(BLCMDnewparticlentuple& r);

	G4String commandName() { return "newparticlentuple"; }

	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	void defineNamedArgs();

	/// PreUserTrackingAction() from BLManager::TrackingAction.
	void PreUserTrackingAction(const G4Track *track);

	/// PostUserTrackingAction() from BLManager::TrackingAction.
	void PostUserTrackingAction(const G4Track *track) { }
};

BLCMDnewparticlentuple defaultNewParticleNTuple;


BLCMDnewparticlentuple::BLCMDnewparticlentuple() : BLCommand(),
						BLManager::TrackingAction()
{
	registerCommand(BLCMDTYPE_DATA);
	setSynopsis("NTuple containing particle tracks when created.");
	setDescription("Note that initial beam particles are included, unless "
		"require is set to 'Parent>0'. "
		"The NTuple contains the same data as a virtualdetector:\n"
		"    x,y,z (mm)\n"
		"    Px,Py,Pz (MeV/c)\n"
		"    t (ns)\n"
		"    PDGid (11=e-, 13=mu-, 22=gamma, 211=pi+, 2212=proton, ...)\n"
		"    EvNum (inexact above 16,777,215)\n"
		"    TrkId (1 is first track of event)\n"
		"    Parent (0 => primary particle)\n"
		"    wt (weight, defaults to 1.0)\n\n"
		"Valid Formats (ignore case): ");
	description += BLNTuple::getFormatList(); 

	name = "";
	format = "";
	filename = "";
	require = "";
	coordinates = "Centerline";
	coordinateType = BLCOORD_CENTERLINE;
	ntuple = 0;
	eval = 0;
}

BLCMDnewparticlentuple::BLCMDnewparticlentuple(BLCMDnewparticlentuple& r) : BLCommand(r), 
						BLManager::TrackingAction(r)
{
	name = r.name;
	format = r.format;
	filename = r.filename;
	require = r.require;
	coordinates = r.coordinates;
	coordinateType = r.coordinateType;
	ntuple = 0;
	eval = 0;
}

int BLCMDnewparticlentuple::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	if(argv.size() != 1) {
		printError("Invalid newparticlentuple command -- need name");
		return -1;
	}

	if(argv[0] == "default") {
		handleNamedArgs(namedArgs);
		return -1;
	}

	BLCMDnewparticlentuple *t = new BLCMDnewparticlentuple(defaultNewParticleNTuple);
	int retval = t->handleNamedArgs(namedArgs);

	t->name = argv[0];
	t->ntuple = BLNTuple::create(t->format,"NTuple",t->name,TrackFields,
							t->filename);

	t->coordinateType = BLCoordinates::getCoordinateType(t->coordinates);
	if(t->require != "") t->eval = new BLEvaluator();

	t->print(argv[0]);

	BLManager::getObject()->registerTrackingAction(t);

	return retval;
}

void BLCMDnewparticlentuple::defineNamedArgs()
{
	argString(format,"format","The NTuple format (see above for list).");
	argString(filename,"filename","The filename for the NTuple.");
	argString(filename,"file","Synonym for filename.");
	argString(require,"require","Expression which must be nonzero to include the track (default=1)",false);
	argString(coordinates,"coordinates","Coordinates: global, centerline, or reference (default=c).");
}

void BLCMDnewparticlentuple::PreUserTrackingAction(const G4Track *track)
{
	// only use reference coordinates when they are valid
	BLManagerState state = BLManager::getObject()->getState();
	if(coordinateType == BLCOORD_REFERENCE && state != BEAM) return;

	if(eval) {
		eval->setTrackVariables(track,coordinateType);
		G4double v = eval->evaluate(require);
		if(!eval->isOK()) {
			BLCommand::printError("newparticlentuple: invalid expression require='%s'\n",
				require.c_str());
			delete eval;
			eval = 0;
			v = 1.0;
		}
		if(v == 0.0) return;
	}

	G4RunManager* runmgr = G4RunManager::GetRunManager();
	const G4Event* event = runmgr->GetCurrentEvent();
	int evId = event->GetEventID();
	G4ThreeVector position = track->GetPosition();
	G4double time = track->GetGlobalTime();
	G4ThreeVector momentum = track->GetMomentum();

	// transform to desired coordinates, if available
	BLCoordinates *c = (BLCoordinates *)track->GetUserInformation();
	if(c && c->isValid()) {
		c->getCoords(coordinateType,position);
		momentum = c->getRotation() * momentum;
	}

	float data[12];
	data[0] = position[0]/mm;		// x (mm)
	data[1] = position[1]/mm;		// y (mm)
	data[2] = position[2]/mm;		// z (mm)
	data[3] = momentum[0]/MeV;		// Px (MeV/c)
	data[4] = momentum[1]/MeV;		// Py (MeV/c)
	data[5] = momentum[2]/MeV;		// Pz (MeV/c)
	data[6] = time/ns;			// t (ns)
	data[7] = track->GetDefinition()->GetPDGEncoding();
	data[8] = evId;				// Event ID
	data[9] = 
		BLManager::getObject()->getExternalTrackID(track->GetTrackID());
	data[10] = track->GetParentID();	// Parent ID
	data[11] = track->GetWeight();		// Weight

	if(ntuple) ntuple->appendRow(data,NTrackFields);
}

