//	BLCMDntuple.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/
//
//	NOTE: this file is #included by BLCMDvirtualdetector.cc in order
//	to define the BLCMDntuple class so it can call setDetector() and 
//	appendRow(). It pre-defines INCLUDE_ONLY.

#define _GNU_SOURCE 1
#include <string.h>
#include <vector>
#include <fnmatch.h>
#include <assert.h>

#include "BLNTuple.hh"
#include "BLCommand.hh"
#include "BLManager.hh"

typedef unsigned long long Bitmask;
const Bitmask Bit = 1;
const int MAX_DETECTORS = sizeof(Bitmask)*8;

/**	class BLCMDntuple - implements an ntuple.
 *
 *	An ntuple holds the data from multiple virtualdetector-s in a single 
 *	NTuple, with one row per event. If multiple tracks hit a given
 *	virtual detector in the ntuple, only the first one is kept.
 **/
class BLCMDntuple : public BLCommand, public BLManager::EventAction {
	G4String name;
	G4String category;
	G4String detectors;	// comma-separated list: detector patterns
	G4String required;	// comma-separated list: required det patterns
	G4String format;
	G4int minHit;
	// internal data:
	G4String fields;	// colon-separated list: fields
	BLNTuple *ntuple;
	float *data;
	int ndata;
	std::map<void*,int> place;	// map from caller to index
	std::vector<int> start;		// index in data[]
	Bitmask need;
	Bitmask got;
	int nHit;
	static std::vector<BLCMDntuple*> list;
	static G4String possibilities;
public:
	/// Default constructor.
	BLCMDntuple();

	/// Destructor.
	virtual ~BLCMDntuple() { }

	/// Copy constructor.
	BLCMDntuple(const BLCMDntuple& r);

	/// commandName() returns "ntuple".
	G4String commandName() { return "ntuple"; }

	/// command() implements the ntuple command.
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	/// defineNamedArgs() defines the named arguments for the command.
	void defineNamedArgs();

	/// addDetector will add a detector to this object.
	/// Returns true if the detector was added.
	bool addDetector(void *caller, G4String name, G4String detFields);

	/// setupFirstEvent() will setup the NTuple (called at first event)
	void setupFirstEvent();

	///  from BLManager::EventAction
	void BeginOfEventAction(const G4Event* event);
	void EndOfEventAction(const G4Event* event);

	/// setDetector is called by BLCMDvirtualdetector as each instance
	/// is placed. Returns # NTuples this virtualdetector is part of.
	/// Called by construct() of each BLCMDvirtualdetector.
	static int setDetector(void *caller, G4String name, G4String detFields);

	/// appendRow is called by BLCMDvirtualdetector as each virtualdetector
	/// is hit by a track.
	/// Called by UserSteppingAction() of each BLCMDvirtualdetector when
	/// a track enters its physical volume.
	static void appendRow(void *caller, float *data, int ndata);

	/// returns true if the string s matches one of the patterns in pattern.
	static bool matches(G4String s, G4String pattern);
};

#ifndef INCLUDE_ONLY

BLCMDntuple defaultNtuple;
std::vector<BLCMDntuple*> BLCMDntuple::list;
G4String BLCMDntuple::possibilities;

BLCMDntuple::BLCMDntuple() : BLCommand(), BLManager::EventAction(),
						place(), start()
{
	registerCommand(BLCMDTYPE_DATA);
	setSynopsis("Define a per-event NTuple containing multiple detectors.");
	setDescription("An ntuple holds the data from multiple "
		"virtualdetector-s in a single NTuple, with one row per event. "
		"Up to 64 virtualdetector-s can be used. "
		"If multiple tracks hit a given virtual detector during the "
		"event, only the first one is kept in this ntuple. "
		"virtualdetector-s are specified by patterns identical to "
		"UNIX file-matching, so '*Det*' matches any virtualdetector "
		"with 'Det' anywhere in its name, etc. The 'required' argument "
		"permits events to be omitted unless all of the matching "
		"virtualdetector-s were hit at least once in the event. The "
		"patterns in 'required' are applied only to virtualdetectors "
		"in the ntuple, so a simple '*' requires a hit in every one.\n"
		"\nNOTE: the name of a virtualdetector is by default the "
		"concatenation of its ancestors' names before its own (except "
		"World), unless "
		"rename=NAME was used in its place command. The patterns are "
		"applied to the names of the virtualdetectors as they were "
		"placed (including rename), not the bare name of the "
		"virtualdetector command. If 'rename=det#' was used to place "
		"the virtualdetector-s, you probably want a * to match the # "
		"or list them individually (det1,det2,det3...).\n\n"
		"Valid Formats (ignore case): ");
	description += BLNTuple::getFormatList(); 
	// initial default values:
	name = "";
	category = "NTuple";
	detectors = "";
	required = "*";
	format = "";
	minHit = 0;
	fields = "";
	ntuple = 0;
	data = 0;
	ndata = 0;
	need = 0;
	got = 0;
	nHit = 0;
}

BLCMDntuple::BLCMDntuple(const BLCMDntuple &r) : BLCommand(r), BLManager::EventAction(),
						place(), start()
{
	name = "";
	category = r.category;
	detectors = r.detectors;
	required = r.required;
	format = r.format;
	minHit = r.minHit;
	fields = r.fields;
	ntuple = 0;
	data = 0;
	ndata = 0;
	need = r.need;
	got = 0;
	nHit = 0;
}


int BLCMDntuple::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	if(argv.size() != 1) {
		printError("Invalid ntuple command -- need name");
		return -1;
	}

	if(argv[0] == "default") {
		handleNamedArgs(namedArgs);
		return -1;
	}

	BLCMDntuple *t = new BLCMDntuple(defaultNtuple);
	int retval = t->handleNamedArgs(namedArgs);

	t->name = argv[0];

	t->print(argv[0]);

	BLManager::getObject()->registerEventAction(t,false);

	list.push_back(t);

	return retval;
}

void BLCMDntuple::defineNamedArgs()
{
	argString(category,"category","The category of the NTuple.");
	argString(detectors,"detectors","A comma-separated list of detector patterns.");
	argString(required,"required","A comma-separated list of required detector patterns(default=*).");
	argString(format,"format","NTuple format (see above for list).");
	argInt(minHit,"minHit","Minimum number of detectors hit (default 0).");
	argString(required,"require","Synonym for required.");
	argInt(minHit,"minHits","Synonym for minHit.");
}

bool BLCMDntuple::addDetector(void *caller, G4String name, G4String detFields)
{
	if(!matches(name,detectors)) return false;

	int j = start.size();
	place[caller] = j;
	start.push_back(ndata);

	if(matches(name,required))
		need |= (Bit<<j);

	unsigned place = 0;
	while(place != detFields.npos) {
		unsigned next = detFields.find(":",place);
		G4String word = detFields.substr(place,
			 (next==detFields.npos ? detFields.npos : next-place));
		if(fields.size() != 0) fields += ":";
		fields += name;
		fields += "_";
		fields += word;
		++ndata;
		place = next;
		if(next != detFields.npos) place = next + 1;
	}

	return true;
}

void BLCMDntuple::setupFirstEvent()
{
	if(ndata <= 0) {
		printError("ntuple: no detectors in ntuple '%s'!  "
			   "Possibilities are:\n%s\n",name.c_str(),
			   possibilities.c_str());
		data = new float[1];
		need = ~(Bitmask)0;
		return;
	}

	data = new float[ndata];
	ntuple = BLNTuple::create(format,category,name,fields,"");
}

int BLCMDntuple::setDetector(void *caller, G4String name, G4String detFields)
{
	int n = 0;

	for(unsigned i=0; i<list.size(); ++i) {
		if(list[i]->addDetector(caller,name,detFields))
			++n;
	}

	if(possibilities.size() > 0) possibilities += ",";
	possibilities += name;

	return n;
}

void BLCMDntuple::appendRow(void *caller, float *detData, int ndetData)
{
	for(unsigned i=0; i<list.size(); ++i) {
		BLCMDntuple *p = list[i];
		if(p->place.count(caller) > 0) {
			unsigned j = p->place[caller];
			assert(j < p->start.size());
			Bitmask me = (Bit<<j);
			if((p->got&me) == 0) {	// keep the first track
				p->got |= me;
				++p->nHit;
				int k = p->start[j];
				for(int l=0; l<ndetData; ++l) {
					assert(k >= 0 && k < p->ndata);
					p->data[k++] = detData[l];
				}
			}
		}
	}
}

void BLCMDntuple::BeginOfEventAction(const G4Event* event)
{
	if(!data)
		setupFirstEvent();

	got = (Bitmask)0;
	nHit = 0;

	for(int i=0; i<ndata; ++i)
		data[i] = 0.0;
}

void BLCMDntuple::EndOfEventAction(const G4Event* event)
{
	if((need&got) == need && nHit >= minHit)
		ntuple->appendRow(data,ndata);
}

bool BLCMDntuple::matches(G4String s, G4String pattern)
{
	unsigned place = 0;

	while(place != pattern.npos) {
		unsigned next = pattern.find(",",place);
		G4String word = pattern.substr(place,
			    (next==pattern.npos ? pattern.npos : next-place));
		if(fnmatch(word.c_str(),s.c_str(),0) == 0)
			return true;
		place = next;
		if(next != pattern.npos) place = next + 1;
	}

	return false;
}

#endif // INCLUDE_ONLY
