//	BLCMDphysics.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/
//

#include <vector>
#include "G4UImanager.hh"
#include "G4UImessenger.hh"
#include "G4MuIonisation.hh"
#include "G4ProcessTable.hh"
#include "G4ParticleTable.hh"
#include "G4StepLimiter.hh"
#include "G4PhysListFactory.hh"

#include "BLPhysics.hh"
#include "BLCommand.hh"
#include "BLManager.hh"

#include "MICEPhysicsList.hh"
#include "LISAPhysicsList.hh"

/**	class BLCMDphysics implements a command to select from the geant4
 *	physics lists of physics processes.
 **/
class BLCMDphysics : public BLPhysics, public BLCommand, public BLCallback {
	G4VUserPhysicsList *physicsList;
	G4int doStochastics;
	G4String disable;
	// G4double minRangeCut; is protected in BLPhysics
	G4int list;
	std::vector<G4String> disableList;
	G4PhysListFactory *physListFactory;
public:
	/// Default Constructor.
	BLCMDphysics();

	/// Destructor.
	~BLCMDphysics() { }
	
	/// setDoStochastics() sets whether or not stochastic processes
	/// are to be enabled.
	void setDoStochastics(BLSetValue value, G4bool quiet=false);

	/// getPhysicsList() returns the G4PhysicsList.
	G4VUserPhysicsList *getPhysicsList() { return physicsList; }

	/// commandName() returns "physics".
	G4String commandName() { return "physics"; }
	
	/// command() implements the command.
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	/// defineNamedArgs() defines the named arguments.
	void defineNamedArgs();

	/// help() provides special help text for the physics command.
	void help(bool detailed);

	/// callback() is from BLCallback, and adds a StepLimiter to the
	/// physics list for every particle.
	void callback(int type);

	/// isLocalPhysicsList() returns true if thsi is a local physics list.
	bool isLocalPhysicsList(G4String name);

	/// getLocalPhysicsList() returns a local physics list.
	G4VUserPhysicsList *getLocalPhysicsList(G4String name);

	/// listPhysicsLists() will list all known physics lists, with synopsis.
	void listPhysicsLists();

	/// printCase() prints the description of a physics list.
	void printCase(G4String name, G4String description);
	
	/// getSynopsis() returns the synopsis for a physics list, by name.
	G4String getSynopsis(G4String listName);
};

BLCMDphysics defaultPhysicsCmd;

BLCMDphysics::BLCMDphysics() : BLPhysics(), BLCommand(), disableList()
{
	registerCommand(BLCMDTYPE_AUX);
	setSynopsis("Defines the physics processes and controls them.");
	setDescription("Exactly one physics command must be present.\n"
		"This command implements the geant4 physics lists\n"
		"of physics processes.\n"
		"The command is 'physics QGSP' for the QGSP set, and\n"
		"similarly for the other sets. With no argument it prints\n"
		"the available physics lists.\n"
		"Note that stochastic processes are always disabled while\n"
		"tracking the tune and reference particles.\n"
		"The only non-stochastic processes are Transportation and\n"
		"ionization energy loss (with fluctuations disabled).\n"
		"For muon beam studies you may want to disable the Decay "
		"process.\n\n"
		"NOTE: this command defines the particles used throughout "
		"the simulation, so this command must come before others "
		"that use particle names.\n\n"
		"The default all-around physics list for HEP is called "
		"'QGSP_BERT'.");

	physicsList = 0;
	doStochastics = 1;
	list = 0;
	disable = "";
	minRangeCut = 1.0 * mm;
	physListFactory = 0;
}

int BLCMDphysics::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	if(physicsList != 0 && argv.size() > 0) {
		printError("physics: Multiple physics commands not allowed!");
		return -1;
	}

	if(argv.size() == 0)
		return -1;

	if(!physListFactory) physListFactory = new G4PhysListFactory();
	if(isLocalPhysicsList(argv[0]))
		physicsList = getLocalPhysicsList(argv[0]);
	if(!physicsList) {
		if(!physListFactory->IsReferencePhysList(argv[0])) {
			printError("physics: Unknown physics list '%s'",
							argv[0].c_str());
			printError("The choices are:");
			listPhysicsLists();
			return -1;
		}
		physicsList = physListFactory->GetReferencePhysList(argv[0]);
	}

	handleNamedArgs(namedArgs);

	disableList = splitString(disable,',',true);

	BLManager::getObject()->registerPhysics(this);

	// now is too early to manage the physics processes; do it in a callback
	BLManager::getObject()->registerCallback(this,0);

	print(argv[0]);

	return 0;
}

void BLCMDphysics::defineNamedArgs()
{
	argString(disable,"disable","Comma-separated list of processes to disable (e.g. 'Decay,msc').");
	argString(disable,"inactivate","Synonym of disable.");
	argString(disable,"deactivate","Synonym of disable.");
	argInt(doStochastics,"doStochastics","Set to zero to disable all stochastic processes.");
	argDouble(minRangeCut,"minRangeCut","Minimum range cut for particle production (1 mm)");
	argInt(list,"list","Nonzero to list the processes (later in output).");
}

void BLCMDphysics::help(bool detailed)
{
	BLCommand::help(detailed);
	if(detailed) {
		printf("\n             ----- PHYSICS LISTS -----\n\n");
		printf("             Further guidance in selecting physics lists is available at:\n"
		"http://geant4.web.cern.ch/geant4/support/physicsLists/referencePL/index.shtml\n\n");
		printf("             The default all-around physics list for HEP is called 'QGSP_BERT'.\n\n");
		printf("             LHEP uses exclusively parameterized modeling.\n");
		printf("             FTF lists use the FRITIOF description of string excitation\n"
		       "                  and fragmentation.\n");
		printf("             QGSP lists use the quark gluon string model\n");
		printf("             QGSC are as QGSP except applying CHIPS modeling for the nuclear\n"
		       "                  de-excitation.\n");
		printf("             QGSP_BERT is like QGSP, but using Geant4 Bertini cascade.\n");
		printf("             QGSP_BIC is like QGSP, but using Geant4 Binary cascade.\n");
		printf("             _EMV suffix indicates a faster but less accurate EM modeling.\n");
		printf("             _HP suffix uses the data driven high precision neutron package\n"
		       "                  (thermal to 20 MeV).\n");
		printf("             _NQE suffix indicates a list for comparison with earlier release.\n");
		printf("\n             List of available physics lists:\n");
		listPhysicsLists();
	}
}

void BLCMDphysics::setDoStochastics(BLSetValue value, G4bool quiet)
{
	if(value == NORMAL) {
		if(doStochastics)
			value = FORCE_ON;
		else
			value = FORCE_OFF;
	}

	G4UImanager* UI = G4UImanager::GetUIpointer();
	G4String cmd;
	char rangeCmd[64];
	switch(value) {
	case FORCE_OFF:
		// turn off delta rays (and other particle production) by
		// setting production cut very high in energy by setting
		// the required range large.
		UI->ApplyCommand("/run/setCut 1 km");
		// Turn off enegry-loss straggling
		UI->ApplyCommand("/process/eLoss/fluct false");
		cmd = "/process/inactivate ";
		if(!quiet) printf("*** NOTE: Stochastic processes are disabled!\n");
		stochasticsEnabled = false;
		break;
	case FORCE_ON:
		// turn on delta rays (and other particle production)
		sprintf(rangeCmd,"/run/setCut %.6g mm",minRangeCut/mm);
		UI->ApplyCommand(rangeCmd);
		// Turn on enegry-loss straggling
		UI->ApplyCommand("/process/eLoss/fluct true");
		cmd = "/process/activate ";
		if(!quiet) printf("Stochastic processes are enabled.\n");
		stochasticsEnabled = true;
		break;
	case NORMAL:
		return;
	}

	// run through list of processes, applying cmd to stochastic ones
	G4ProcessTable *pt = G4ProcessTable::GetProcessTable();
	G4ProcessTable::G4ProcNameVector *nv = pt->GetNameList();
	if(list) printf("List of physics processes:\n");
	for(unsigned int i=0; i<nv->size(); ++i) {
		G4String name = (*nv)[i];
		if(list) printf("\t%s\n",name.c_str());
		if(isStochasticProcess(name))
			UI->ApplyCommand(cmd+name);
	}
	if(list) printf("\n");
	list = 0;	// print list only once

	// now disable all proceses in the disableList
	cmd = "/process/inactivate ";
	for(unsigned i=0; i<disableList.size(); ++i) {
		if(!quiet) printf("*** NOTE: Disabling physics process '%s'\n",
						disableList[i].c_str());
		UI->ApplyCommand(cmd+disableList[i]);
	}
}

void BLCMDphysics::callback(int type)
{
	G4ParticleTable::G4PTblDicIterator *theParticleIterator =
			G4ParticleTable::GetParticleTable()->GetIterator();
	theParticleIterator -> reset();
	while( (*theParticleIterator)() ) {
		G4ParticleDefinition* particle = theParticleIterator->value();
		G4ProcessManager* pmanager = particle -> GetProcessManager();
		if(!pmanager) {
			printf("BLCMDphysics: particle '%s' has no ProcessManager!\n",
				particle->GetParticleName().c_str());
			continue;
		}
		pmanager -> AddProcess(new G4StepLimiter(),-1,-1,3);
	}
}

bool BLCMDphysics::isLocalPhysicsList(G4String name)
{
	if(name == "MICEPhysicsList") return true;
	if(name == "LISAPhysicsList") return true;
	return false;
}

G4VUserPhysicsList *BLCMDphysics::getLocalPhysicsList(G4String name)
{
	if(name == "MICEPhysicsList") return new MICEPhysicsList();
	if(name == "LISAPhysicsList") return new LISAPhysicsList();
	return 0;
}


void BLCMDphysics::listPhysicsLists()
{
	if(!physListFactory) physListFactory = new G4PhysListFactory();
	const std::vector<G4String> v = physListFactory->AvailablePhysLists();
	for(unsigned i=0; i<v.size(); ++i) {
		G4String n = v[i];
		G4String s = getSynopsis(n);
		printCase(n,s);
	}
	printCase("MICEPhysicsList",getSynopsis("MICEPhysicsList"));
	printCase("LISAPhysicsList",getSynopsis("LISAPhysicsList"));
}

void BLCMDphysics::printCase(G4String name, G4String description)
{
	G4String indent1("             "), indent("       ");
	indent += indent1;
	indent1 += name;
	do { indent1 += " "; } while(indent1.size() < indent.size());
	printf("%s",wrapWords(description,indent1,indent).c_str());
}

G4String BLCMDphysics::getSynopsis(G4String listName)
{
#define CASE(NAME,SYNOPSIS)				\
	if(listName == #NAME) return SYNOPSIS;
again:
#ifndef PHYSICS_NO_USE_CASES
	CASE(FTFP,"For calorimetry -- is similar to QGSP for the treatment of the fragmentation, but the string excitation/fragmentation is changed from quark-gluon string model to a diffractive string excitation similar to that in FRITJOF, and the Lund fragmentation functions.");
	CASE(LBE,"For low background experiments (e.g. underground)");
	CASE(LHEP,"For calorimetry -- is the fastest, when it comes to CPU. It uses the LEP and HEP parametrized models for inelastic scattering. The modeling parametrizes the final states individual inelastic reactions, so you will not see resonances, and the detailed secondary angular distributions for O(100MeV) reactions may not be described perfectly. The average quantities will be well described.");
	CASE(LHEP_BERT,"For 'typical' HEP collider detectors and low-energy penetration shielding (< 5 GeV) -- is as LHEP_GN for energies above 3GeV. Below 3 GeV the Bertini cascade is used for nucleon and pion induced reactions. The price to pay is reduced CPU performance. It does not include gamma and electro nuclear physics.");
	CASE(LHEP_PRECO_HP,"For low-energy penetration shielding (< 5 GeV) -- is like LHEP_PRECO, but uses point-wise evaluated cross-section data to model neutron interactions from thermal neutron energies to ~20MeV. This applies to capture, elastic scattering, fission, and inelastic scattering. Again fission fragments are not available. Please use together with G4NDL3.5 or higher.");
	CASE(QGSC,"For calorimetry and high energy physics trackers -- is as QGSP for the initial reaction, but uses chiral invariant phase-space decay (multi-quasmon fragmentation) to model the behavior of the system's fragmentation.");
	CASE(QGSP,"For calorimetry and high energy physics trackers and high-energy and medium-energy production targets -- uses theory driven modeling for the reactions of energetic pions, kaons, and nucleons. It employs quark gluon string model for the 'punch-through' interactions of the projectile with a nucleus, the string excitation cross-sections being calculated in quasi-eikonal approximation. A pre-equilibrium decay model with an extensive evaporation phase to model the behavior of the nucleus 'after the punch'. It uses current best pion cross-section.");
	CASE(QGSP_BERT,"Like QGSP, but using Geant4 Bertini cascade for primary protons, neutrons, pions and Kaons below ~10GeV. In comparison to experimental data we find improved agreement to data compared to QGSP which uses the low energy parameterised (LEP) model for all particles at these energies. The Bertini model produces more secondary neutrons and protons than the LEP model, yielding a better agreement to experimental data.");
	CASE(QGSP_BIC,"Like QGSP, but using Geant4 Binary cascade for primary protons and neutrons with energies below ~10GeV, thus replacing the use of the LEP model for protons and neutrons In comparison to teh LEP model, Binary cascade better describes production of secondary particles produced in interactions of protons and neutrons with nuclei.");
#endif
	CASE(LISAPhysicsList,"Geant4 examples/advanced/cosmicray_charging");
	CASE(MICEPhysicsList,"The current MICE physics list.");

	return "No synopsis available.";
}
