//	BLCMDtrace.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include <set>

#include "globals.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4ParticleDefinition.hh"
#include "G4RunManager.hh"
#include "G4Event.hh"
#include "G4ThreeVector.hh"

#include "BLCoordinates.hh"
#include "BLEvaluator.hh"
#include "BLGlobalField.hh"
#include "BLCommand.hh"
#include "BLParam.hh"
#include "BLManager.hh"
#include "BLRunManager.hh"
#include "BLNTuple.hh"
#include "BLCollectiveComputation.hh"

const char TraceFields[] =
    "x:y:z:Px:Py:Pz:t:PDGid:EventID:TrackID:ParentID:Weight:Bx:By:Bz:Ex:Ey:Ez";
const int NTraceFields = 18;

/**	class TraceNTuple implements an NTuple for tracing a track.
 **/
class TraceNTuple {
	BLNTuple *ntuple;
	BLCoordinateType coordinateType;
	BLManager *manager;
	BLRunManager *runManager;
public:
	/// constructor.
	TraceNTuple(G4String format, G4String name, G4String filename,
					BLCoordinateType _coordinateType);

	/// destructor.
	virtual ~TraceNTuple();

	/// appendTrace() will append a track to the Trace NTuple.
	void appendTrace(G4Track *track);

	/// close() will close the NTuple.
	void close() { if(ntuple) ntuple->close(); }

	/// annotate() will add an annotation to ASCII NTuples.
	void annotate(G4String line) { if(ntuple) ntuple->annotate(line); }
};

/**	class BLCMDtrace implements the trace command.
 *
 **/
class BLCMDtrace : public BLCommand, public BLManager::SteppingAction,
	    public BLManager::TrackingAction, public BLCollectiveComputation {
	int nTrace;
	G4String format;
	G4int oneNTuple;
	G4int primaryOnly;
	G4String filename;
	G4String require;
	G4String coordinates;
	BLCoordinateType coordinateType;
	bool doTrace;
	TraceNTuple *trace;
	BLEvaluator *eval;
	bool init;
	BLManager *manager;
	BLRunManager *runManager;
	static TraceNTuple *allTrace;
	friend class TraceNTuple;
public:
	/// Constructor
	BLCMDtrace();

	BLCMDtrace(const BLCMDtrace &r);

	G4String commandName() { return "trace"; }
	
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	void defineNamedArgs();

	void newTrace(const G4Track *firstTrack);

	/// UserSteppingAction() from BLManager::SteppingAction.
	void UserSteppingAction(const G4Step *step);

	/// PreUserTrackingAction() from BLManager::TrackingAction.
	void PreUserTrackingAction(const G4Track *track);

	/// PostUserTrackingAction() from BLManager::TrackingAction.
	void PostUserTrackingAction(const G4Track *track);

	/// from BLCollectiveComputation
	virtual void beginCollectiveTracking(std::vector<BLTrackData>& v);
	virtual void collectiveStep(std::vector<BLTrackData>& v);
	virtual void endCollectiveTracking(std::vector<BLTrackData>& v);
};
TraceNTuple *BLCMDtrace::allTrace = 0;
BLCMDtrace defaultTrace;

BLCMDtrace::BLCMDtrace() : BLCommand(), BLManager::SteppingAction(), 
            BLManager::TrackingAction(), BLCollectiveComputation(),
	    format()
{
	registerCommand(BLCMDTYPE_DATA);
	setSynopsis("Specifies tracing of tracks.");
	setDescription("Generates a separate NTuple for each track, with 1\n"
		"row per step, unless oneNTuple is nonzero (in which case "
		"all tracks are put into a single NTuple).\n"
		"So format=ascii generates one file per track with names\n"
		"generated by the pattern in filename (first %d is replaced "
		"by event #, second %d is replaced by trackId); for "
		"oneNTuple, the default filename is AllTracks.txt.\n"
		"\nNote that without a trace command no traces are generated, "
		"so to trace just the tune and reference particles include "
		"a trace command with no arguments.\n\n"
		"In collective tracking mode, oneNTuple must be nonzero, and "
		"the entries will be generated only at collective steps "
		"(usually at a specified deltaT).\n\n"
		"Valid Formats (ignore case): ");
	description += BLNTuple::getFormatList(); 
	nTrace = 0;
	format = "root";
	oneNTuple = 0;
	primaryOnly = 0;
	filename = "";
	require = "";
	coordinates = "Centerline";
	coordinateType = BLCOORD_CENTERLINE;
	doTrace = true;
	trace = 0;
	eval = 0;
	init = false;
	manager = 0;	// too early to set these -- see command()
	runManager = 0;
}

BLCMDtrace::BLCMDtrace(const BLCMDtrace& r) : BLCommand(r), 
	    BLManager::SteppingAction(), BLManager::TrackingAction(),
	    BLCollectiveComputation(), format()
{
	nTrace = r.nTrace;
	format = r.format;
	oneNTuple = r.oneNTuple;
	primaryOnly = r.primaryOnly;
	filename = r.filename;
	require = r.require;
	coordinates = r.coordinates;
	coordinateType = r.coordinateType;
	doTrace = r.doTrace;
	trace = r.trace;
	eval = 0;
	init = r.init;
	manager = r.manager;
	runManager = r.runManager;
}

int BLCMDtrace::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	manager = BLManager::getObject();
	runManager = BLRunManager::getObject();

	int retval = handleNamedArgs(namedArgs);

	if(!init) {
		init = true;
		manager->registerSteppingAction(this);
		manager->registerTrackingAction(this);
		runManager->registerCollectiveComputation(this);
	}

	if(filename == "")
		filename = (oneNTuple!=0 ? "AllTracks" : "Ev%dTrk%d");

	coordinateType = BLCoordinates::getCoordinateType(coordinates);
	if(require != "") eval = new BLEvaluator();

	print("");

	return retval;
}

void BLCMDtrace::defineNamedArgs()
{
	argInt(nTrace,"nTrace","Number of tracks to trace.");
	argString(format,"format","Format of the NTuple (see above for list).");
	argInt(oneNTuple,"oneNTuple","Nonzero to put all traces into a single NTuple.");
	argInt(primaryOnly,"primaryOnly","Nonzero to trace only primary tracks.");
	argString(filename,"filename","Filename (Ev%dTrk%d.txt or AllTracks.txt).");
	argString(filename,"file","synonym for filename.");
	argString(require,"require","Expression which must be nonzero to trace the track (default=1)",false);
	argString(coordinates,"coordinates","Coordinates: global, centerline, or reference (default=c).");
}

void BLCMDtrace::newTrace(const G4Track *firstTrack)
{
	static int prevEvNum = -999;

	int evNum = G4RunManager::GetRunManager()->GetCurrentEvent()->GetEventID();
	int trkId = manager->getExternalTrackID(firstTrack->GetTrackID());

	if(oneNTuple != 0) {
		if(!allTrace) allTrace = new TraceNTuple(format,"AllTracks",
						filename, coordinateType);
	} else {
		if(trace) trace->close();
	}
	trace = 0;

	if(eval) {
		eval->setTrackVariables(firstTrack,coordinateType);
		G4double v = eval->evaluate(require);
		if(!eval->isOK()) {
			BLCommand::printError("trace: invalid expression require='%s'\n",
				require.c_str());
			delete eval;
			eval = 0;
			v = 1.0;
		}
		if(v == 0.0) return;
	}

	if(oneNTuple != 0) {
		trace = allTrace;
	} else if(evNum < 0) {
		if(evNum != prevEvNum) {
			prevEvNum = evNum;
			char tmp[128];
			if(evNum == -2)
				trace = new TraceNTuple(format,"TuneParticle",
					"TuneParticle",coordinateType);
			else if(evNum == -1)
				trace = new TraceNTuple(format,"ReferenceParticle",
					"ReferenceParticle",coordinateType);
			else
				trace = 0;
		} 
	} else {
		char tmp[128];
		sprintf(tmp,filename.c_str(),evNum,trkId);
		trace = new TraceNTuple(format,tmp,tmp,coordinateType);
	}
}


void BLCMDtrace::UserSteppingAction(const G4Step *step)
{
	if(manager->getState() == SPECIAL || runManager->getCollectiveMode())
		return;

	if(!trace || !doTrace) return;

	G4Track *track = step->GetTrack();
	trace->appendTrace(track);
}

void BLCMDtrace::PreUserTrackingAction(const G4Track *track)
{
	int evNum = runManager->GetCurrentEvent()->GetEventID();

	doTrace = false;

	if(evNum < 0) {
		// cannot use reference coordinates for tune or reference
		doTrace = (coordinateType != BLCOORD_REFERENCE);
		if(doTrace) newTrace(track);
	} else if(runManager->getCollectiveMode() && nTrace > 0) {
		if(oneNTuple == 0) {
			G4Exception("trace","Collective mode, oneNTuple=0",
				FatalException,"oneNTuple must be nonzero");
		}
		newTrace(track);
	} else if(nTrace > 0) {
		doTrace = ((primaryOnly == 0) || (track->GetParentID() == 0));
		if(doTrace) {
			--nTrace;
			newTrace(track);
		}
	} else {
		if(trace) {
			trace->close();
			trace = 0;
		}
	}

	if(trace && doTrace && 
			!runManager->getCollectiveMode()) {
		char tmp[64];
		sprintf(tmp,"# Event %d Track %d",evNum,manager->
				getExternalTrackID(track->GetTrackID()));
		trace->annotate(tmp);
	}
}

void BLCMDtrace::PostUserTrackingAction(const G4Track *track)
{
	if(!trace) return;

	if(doTrace && !runManager->getCollectiveMode())
		trace->annotate("");

	int evNum = G4RunManager::GetRunManager()->GetCurrentEvent()->
								GetEventID();
	if(oneNTuple == 0 && evNum >= 0) {
		trace->close();
		trace = 0;
	}
}

void BLCMDtrace::beginCollectiveTracking(std::vector<BLTrackData>& v)
{
	(void)v;
}

void BLCMDtrace::collectiveStep(std::vector<BLTrackData>& v)
{
	if(trace) {
		for(unsigned i=0; i<v.size(); ++i) {
			G4Track *track = v[i].track;
			G4Event *event = v[i].event;
			runManager->setCurrentEvent(event);
			G4TrackStatus trackStatus = track->GetTrackStatus();
			if(trackStatus != fAlive && 
						trackStatus != fStopButAlive)
				continue;
			if(primaryOnly != 0 && track->GetParentID() > 0)
				continue;
			trace->appendTrace(track);
		}
	}
}

void BLCMDtrace::endCollectiveTracking(std::vector<BLTrackData>& v)
{
	(void)v;
}



TraceNTuple::TraceNTuple(G4String format, G4String name, G4String filename,
					BLCoordinateType _coordinateType)
{
	ntuple = BLNTuple::create(format,"Trace",name,TraceFields,filename);
	coordinateType = _coordinateType;
	manager = BLManager::getObject();
	runManager = BLRunManager::getObject();
}

TraceNTuple::~TraceNTuple()
{
	if(ntuple) delete ntuple;
	ntuple = 0;
}

void TraceNTuple::appendTrace(G4Track *track)
{
	const G4Event* event = runManager->GetCurrentEvent();
	int evNum = event->GetEventID();
	G4ThreeVector position = track->GetPosition();
	G4double time = track->GetGlobalTime();
	G4ThreeVector momentum = track->GetMomentum();

	G4double point[4], field[6];
	point[0] = position[0];
	point[1] = position[1];
	point[2] = position[2];
	point[3] = time;
	BLGlobalField::getObject()->GetFieldValue(point,field);
	G4ThreeVector B(field[0],field[1],field[2]);
	G4ThreeVector E(field[3],field[4],field[5]);

	// transform to desired coordinates, if appropriate
	BLCoordinates *c = (BLCoordinates *)track->GetUserInformation();
	if(c && c->isValid()) {
		BLManagerState state = manager->getState();
		c->getCoords((state==BEAM?coordinateType:BLCOORD_CENTERLINE),
								position);
		momentum = c->getRotation() * momentum;
		B = c->getRotation() * B;
		E = c->getRotation() * E;
	}

	float data[NTraceFields];
	data[0] = position[0]/mm;		// x (mm)
	data[1] = position[1]/mm;		// y (mm)
	data[2] = position[2]/mm;		// z (mm)
	data[3] = momentum[0]/MeV;		// Px (MeV/c)
	data[4] = momentum[1]/MeV;		// Py (MeV/c)
	data[5] = momentum[2]/MeV;		// Pz (MeV/c)
	data[6] = time/ns;			// t (ns)
	data[7] = track->GetDefinition()->GetPDGEncoding();
	data[8] = evNum;				// Event ID
	data[9] = manager->getExternalTrackID(track->GetTrackID());
	data[10] = track->GetParentID();	// Parent ID
	data[11] = track->GetWeight();		// Weight
	data[12] = B[0]/tesla;			// Bx
	data[13] = B[1]/tesla;			// By
	data[14] = B[2]/tesla;			// Bz
	data[15] = E[0]/(megavolt/meter);	// Ex
	data[16] = E[1]/(megavolt/meter);	// Ey
	data[17] = E[2]/(megavolt/meter);	// Ez

	ntuple->appendRow(data,NTraceFields);
}
