//	BLCMDzntuple.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>

#include "BLCommand.hh"
#include "BLManager.hh"
#include "BLParam.hh"
#include "BLCoordinates.hh"
#include "BLNTuple.hh"
#include "BLGlobalField.hh"
#include "BLEvaluator.hh"

// include the class definition only
#define INCLUDE_ONLY
#include "BLCMDntuple.cc"

#ifdef WIN32
#define snprintf _snprintf
#endif
#ifdef __CYGWIN__
#include "mysnprintf.hh"
#endif


const char TrackFields[] =
    "x:y:z:Px:Py:Pz:t:PDGid:EventID:TrackID:ParentID:Weight";
const unsigned NTrackFields = 12;

/**	class BLCMDzntuple will generate an NTuple for each of a list of Z
 *	positions.
 **/
class BLCMDzntuple : public BLCommand, public BLCallback {
	struct Entry : public BLManager::ZSteppingAction {
		BLCMDzntuple *zntuple;
		BLNTuple *ntuple;
		BLEvaluator *eval;
		BLCoordinateType coordinateType;
		int nNTuples;	// # ntuples this virtualdetector is part of
		Entry(BLCMDzntuple *znt, G4double zpos, 
					BLCoordinateType _coordinateType);
		void UserZSteppingAction(const G4Track *track);
	};
	G4String z;
	G4String zloop;
	G4int noSingles;
	G4String format;
	G4String file;
	G4String require;
	G4int referenceParticle;
	G4String coordinates;
	BLCoordinateType coordinateType;
public:
	/// Constructor.
	BLCMDzntuple();
	BLCMDzntuple(BLCMDzntuple& r);

	/// commandName() returns "zntuple".
	virtual G4String commandName() { return "zntuple"; }
	
	/// command() implements the zntuple command.
	virtual int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	/// defineNamedArgs() defines the named arguments for this command.
	virtual void defineNamedArgs();

	/// callback from BLCallback.
	virtual void callback(int type);
};

BLCMDzntuple defaultZntuple;

BLCMDzntuple::BLCMDzntuple() : BLCommand(), BLCallback() 
{
	registerCommand(BLCMDTYPE_DATA);
	setSynopsis("Generate an NTuple for each of a list of Z positions.");
	setDescription("Generates an NTuple like a virtualdetector, but\n"
			"without a physical volume. Tracks are forced to \n"
			"take steps within 2mm surrounding each desired z\n"
			"position, and they are interpolated to the desired\n"
			"z position. Each z position generates a separate\n"
			"NTuple named Z123 (etc.). z accepts a list of z\n"
			"positions, and zloop can generate a set of equally\n"
			"spaced z positions.\n\n"
			"Each value in z and zloop can be an expression using "
			"double constants and the usual C operators and "
			"functions.\n\n"
			"Valid Formats (ignore case): ");
	description += BLNTuple::getFormatList(); 

	z = "";
	zloop = "";
	noSingles = 0;
	format = "root";
	file = "";
	require = "";
	referenceParticle = 0;
	coordinates = "Centerline";
	coordinateType = BLCOORD_CENTERLINE;
}

BLCMDzntuple::BLCMDzntuple(BLCMDzntuple& r) : BLCommand(r), BLCallback(r)
{
	z = r.z;
	zloop = r.zloop;
	noSingles = r.noSingles;
	format = r.format;
	file = r.file;
	require = r.require;
	referenceParticle = r.referenceParticle;
	coordinates = r.coordinates;
	coordinateType = r.coordinateType;
}

int BLCMDzntuple::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	BLCMDzntuple *p = new BLCMDzntuple(defaultZntuple);

	int retval = p->handleNamedArgs(namedArgs);

	p->coordinateType = BLCoordinates::getCoordinateType(p->coordinates);

	p->print("");

	// must defer creation of BLNTuple-s
	BLManager::getObject()->registerCallback(p,0);

	return retval;
}

void BLCMDzntuple::defineNamedArgs()
{
	argString(z,"z","Comma-separated list of Z positions (mm)");
	argString(zloop,"zloop","Loop in z, first:last:incr (mm)");
	argInt(noSingles,"noSingles","Set to 1 to omit the NTuple for singles.");
	argString(format,"format","NTuple format (see above for list)");
	argString(file,"file","Output filename ('' uses name to determine filename)");
	argString(file,"filename","Synonym for file");
	argString(require,"require","Expression which must be nonzero to include the track (default=1)",false);
	argInt(referenceParticle,"referenceParticle","Set to 1 to include the Reference Particle.");
	argString(coordinates,"coordinates","Coordinates: global, centerline, or reference (default=c).");
}

void BLCMDzntuple::callback(int type)
{
	if(type != 0) return;

	// register the z position(s) in the string z
	std::vector<G4double> vect = getList(z,",");
	if(vect.size() > 0) {
		for(unsigned i=0; i<vect.size(); ++i) {
			Entry *e = new Entry(this,vect[i],coordinateType);
			int when = 4;
			if(referenceParticle != 0 || 
			   (e->ntuple && e->ntuple->needsReference()))
				when |= 2;
			BLManager::getObject()->registerZStep(vect[i],e,when);
		}
	} else if(z.size() > 0) {
		printError("printf: Syntax error in z");
	}

	// register the z position(s) in the string zloop
	vect = getList(zloop,",:");
	if(vect.size() == 3) {
		if(vect[2] < 1.0*mm || vect[0] > vect[1]) {
			printError("printf: invalid zloop '%s'",zloop.c_str());
		} else {
			while(vect[0] <= vect[1]) {
			    Entry *e = new Entry(this,vect[0],coordinateType);
			    int when = 4;
			    if(referenceParticle != 0 || 
			       (e->ntuple && e->ntuple->needsReference()))
			       	when |= 2;
			    BLManager::getObject()->registerZStep(vect[0],
								e,when);
			    vect[0] += vect[2];
			}
		}
	} else if(zloop.size() > 0) {
		printError("printf: invalid zloop '%s'",zloop.c_str());
	}
}

BLCMDzntuple::Entry::Entry(BLCMDzntuple *znt, G4double zpos, 
					BLCoordinateType _coordinateType)
{
	char name[128];
	snprintf(name,sizeof(name),"Z%.0f",floor(zpos+0.5));

	zntuple = znt;
	if(zntuple->noSingles == 0)
		ntuple = BLNTuple::create(zntuple->format,"NTuple",name,
						TrackFields,zntuple->file);
	else
		ntuple = 0;

	eval = 0;
	if(zntuple->require != "") eval = new BLEvaluator();
	coordinateType = _coordinateType;

	// define this virtualdetector to the ntuple command
	nNTuples = BLCMDntuple::setDetector(this,name,TrackFields);
}

void BLCMDzntuple::Entry::UserZSteppingAction(const G4Track *track)
{
	// only use reference coordinates when they are valid
	BLManagerState state = BLManager::getObject()->getState();
	if(coordinateType == BLCOORD_REFERENCE && state != BEAM) return;

	if(eval) {
		eval->setTrackVariables(track,coordinateType);
		G4double v = eval->evaluate(zntuple->require);
		if(!eval->isOK()) {
			BLCommand::printError("beamlossntuple: invalid expression require='%s'\n",
				zntuple->require.c_str());
			delete eval;
			eval = 0;
			v = 1.0;
		}
		if(v == 0.0) return;
	}

	G4RunManager* runmgr = G4RunManager::GetRunManager();
	const G4Event* event = runmgr->GetCurrentEvent();
	int evId = event->GetEventID();
	G4ThreeVector position = track->GetPosition();
	G4double time = track->GetGlobalTime();
	G4ThreeVector momentum = track->GetMomentum();

	// get # data fields, and get B and E if needed
	unsigned ndata = (ntuple ? ntuple->getNData() : NTrackFields);
	G4double field[6];
	if(ndata > NTrackFields) {
		G4double point[4];
		point[0] = position[0];
		point[1] = position[1];
		point[2] = position[2];
		point[3] = time;
		BLGlobalField::getObject()->GetFieldValue(point,field);
	}

	// transform to desired coordinates, if available
	BLCoordinates *c = (BLCoordinates *)track->GetUserInformation();
	if(c && c->isValid()) {
		c->getCoords(coordinateType,position);
		momentum = c->getRotation() * momentum;
	}


	float data[18];
	assert(ndata <= sizeof(data)/sizeof(data[0]));
	data[0] = position[0]/mm;		// x (mm)
	data[1] = position[1]/mm;		// y (mm)
	data[2] = position[2]/mm;		// z (mm)
	data[3] = momentum[0]/MeV;		// Px (MeV/c)
	data[4] = momentum[1]/MeV;		// Py (MeV/c)
	data[5] = momentum[2]/MeV;		// Pz (MeV/c)
	data[6] = time/ns;			// t (ns)
	data[7] = track->GetDefinition()->GetPDGEncoding();
	data[8] = evId;				// Event ID
	data[9] = 
		BLManager::getObject()->getExternalTrackID(track->GetTrackID());
	data[10] = track->GetParentID();	// Parent ID
	data[11] = track->GetWeight();		// Weight
	if(ndata > NTrackFields) {
		G4ThreeVector B(field[0],field[1],field[2]);
		G4ThreeVector E(field[3],field[4],field[5]);
		if(c && c->isValid()) {
			B = c->getRotation() * B;
			E = c->getRotation() * E;
		}
		data[12] = B[0]/tesla;
		data[13] = B[1]/tesla;
		data[14] = B[2]/tesla;
		data[15] = E[0]/(megavolt/meter);
		data[16] = E[1]/(megavolt/meter);
		data[17] = E[2]/(megavolt/meter);
	}

	if(ntuple) ntuple->appendRow(data,ndata);

	// insert data into BLCMDntuple-s
	if(nNTuples > 0) BLCMDntuple::appendRow(this,data,NTrackFields);
}
