//	BLNTuple.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#include <vector>
#include <time.h>
#include <assert.h>

#include "BLNTuple.hh"
#include "BLParam.hh"
#include "BLTrackFile.hh"
#include "BLFOR009.hh"
#include "BLCommand.hh"
#include "BLLoad.hh"


bool BLNTuple::enableOutput = true;
G4String BLNTuple::defaultType = "root";
std::map<G4String,BLNTupleHandler*> *BLNTuple::handler = 0;
std::vector<BLNTuple *> BLNTuple::list;



void BLNTuple::registerHandler(G4String typeName, BLNTupleHandler *h)
{
	// permit this to be called during initializations, without
	// worrying about ordering.
	if(!handler)
		handler = new std::map<G4String,BLNTupleHandler*>();

	for(unsigned i=0; i<typeName.size(); ++i)
		typeName[i] = tolower(typeName[i]);

	(*handler)[typeName] = h;
}

BLNTupleHandler *BLNTuple::getHandler(G4String type)
{
	for(unsigned i=0; i<type.size(); ++i)
		type[i] = tolower(type[i]);

	if(!handler) return 0;
	return (*handler)[type]; // returns 0 for unknown type
}

BLNTuple *BLNTuple::create(G4String type, G4String category, G4String name,
					G4String fields, G4String filename)
{
	if(type == "") type = defaultType;
	if(!enableOutput) type = "Dummy";

	BLNTupleHandler *h = getHandler(type);
	if(h) {
		BLNTuple *p = h->create(type,category,name,fields,filename);
		if(p) list.push_back(p);
		return p;
	}

	BLCommand::printError("BLNTuple: unknown format '%s'\n",
					type.c_str());

	return 0;
}

BLNTuple *BLNTuple::read(G4String type, G4String category, G4String name,
					G4String fields, G4String filename)
{
	if(type == "") type = defaultType;

	BLNTupleHandler *h = getHandler(type);
	if(h) {
		BLNTuple *p = h->read(type,category,name,fields,filename);
		return p;
	}

	BLCommand::printError("BLNTuple: unknown format '%s'\n",
					type.c_str());

	return 0;
}

void BLNTuple::flushAll()
{
	for(unsigned i=0; i<list.size(); ++i)
		list[i]->flush();
}

void BLNTuple::closeAll()
{
	for(unsigned i=0; i<list.size(); ++i)
		list[i]->close();
}

void BLNTuple::summary()
{
	for(unsigned i=0; i<list.size(); ++i)
		list[i]->doSummary();
}

void BLNTuple::update()
{
	// no-op for now....
}

G4String BLNTuple::getFormatList()
{
	G4String ret;
	if(handler == 0) return ret;
	std::map<G4String,BLNTupleHandler*>::iterator i;
	for(i=(*handler).begin(); i!=(*handler).end(); ++i) {
		if(i->first == "Dummy") continue;
		if(i != (*handler).begin()) ret += " ";
		ret += i->first;
	}
	return ret;
}


#ifdef USE_SHARED_OBJECTS

// auto-loading handler for HistoScope NTuples
class HistoHandlerLoader : public BLNTupleHandler {
	bool loadHistoSharedObject(G4String type) {
		if(!BLLoad::load("g4bl-histoscope")) {
		    BLCommand::printError("NTuple: HistoScope not "
				"supported");
		    return false;
		}
		// verify the loaded .so registered this type
		if(BLNTuple::getHandler(type) == this) {
		    BLCommand::printError("NTuple: HistoScope "
				"INTERNAL ERROR");
		    return false;
		}
		return true;
	}
public:
	HistoHandlerLoader() {
		if(BLNTuple::getHandler("histo") == 0) {
			BLNTuple::registerHandler("histo",this); 
			BLNTuple::registerHandler("HistoScope",this); 
			BLNTuple::registerHandler("histoscope",this); 
		}
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		if(!loadHistoSharedObject(type)) return 0;
		return BLNTuple::getHandler(type)->create(type,category,name,
						fields,filename);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields, G4String filename) {
		if(!loadHistoSharedObject(type)) return 0;
		return BLNTuple::getHandler(type)->read(type,category,name,
						fields, filename);
	}
};
HistoHandlerLoader defaultHistoHandlerLoader;


// auto-loading handler for Root NTuples
class RootHandlerLoader : public BLNTupleHandler {
	bool loadRootSharedObject(G4String type) {
		if(!BLLoad::load("g4bl-root")) {
		    BLCommand::printError("NTuple: Root not "
				"supported");
		    return false;
		}
		// verify the loaded .so registered this type
		if(BLNTuple::getHandler(type) == this) {
		    BLCommand::printError("NTuple: Root "
				"INTERNAL ERROR");
		    return false;
		}
		return true;
	}
public:
	RootHandlerLoader() {
		if(BLNTuple::getHandler("root") == 0) {
			BLNTuple::registerHandler("root",this); 
			BLNTuple::registerHandler("Root",this); 
		}
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		if(!loadRootSharedObject(type)) return 0;
		return BLNTuple::getHandler(type)->create(type,category,name,
						fields,filename);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields, G4String filename) {
		if(!loadRootSharedObject(type)) return 0;
		return BLNTuple::getHandler(type)->read(type,category,name,
						fields,filename);
	}
};
RootHandlerLoader defaultRootHandlerLoader;

#endif // USE_SHARED_OBJECTS



/**	class DummyNTuple implements a Dummy NTuple.
 **/
class DummyNTuple : public BLNTuple {
public:
	DummyNTuple() { }
	~DummyNTuple() { }
	virtual void appendRow(float data[], int n) { }
	virtual bool readRow(float data[], int ndata) { return false; }
	virtual int getNData() { return 1; }
	void flush() { }
	void close() { }
	void doSummary() { }
};

class DummyNTupleHandler : public BLNTupleHandler {
public:
	DummyNTupleHandler() { 
		BLNTuple::registerHandler("Dummy",this);
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		return new DummyNTuple();
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields,G4String filename) {
		return 0;
	}
};
DummyNTupleHandler defaultDummyNTupleHandler;



/**	class TrackFileNTuple implements a BLTrackFile NTuple.
 **/
const int NTrackFileFields=12;
class TrackFileNTuple : public BLNTuple {
	BLTrackFile *tf;
	G4String title;
	int entries;
	static std::vector<TrackFileNTuple*> list;
public:
	TrackFileNTuple(G4String category, G4String name, G4String fields,
							G4String filename);
	~TrackFileNTuple() { close(); }
	virtual void appendRow(float data[], int n);
	virtual bool readRow(float data[], int ndata) { return false; }
	virtual int getNData() { return NTrackFileFields; }
	void flush() { if(tf) tf->flush(); }
	void close() { if(tf) delete tf; tf = 0; }
	void doSummary() {
		printf("NTuple %-15s %8d entries\n",title.c_str(), entries);
	}
};
std::vector<TrackFileNTuple*> TrackFileNTuple::list;

class TrackFileNTupleHandler : public BLNTupleHandler {
public:
	TrackFileNTupleHandler() { 
		BLNTuple::registerHandler("TrackFile",this);
		BLNTuple::registerHandler("BLTrackFile",this);
		BLNTuple::registerHandler("trackfile",this);
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		return new TrackFileNTuple(category,name,fields,filename);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields,G4String filename) {
		return 0;
	}
};
TrackFileNTupleHandler defaultTrackFileNTupleHandler;

TrackFileNTuple::TrackFileNTuple(G4String category, G4String name,
					G4String fields, G4String filename)
{
	if(filename == "") filename = name+".txt";
	if(filename.find(".txt") == filename.npos)
		filename += ".txt";
	tf = new BLTrackFile(filename,category+"/"+name,"w");
	title = name;
	entries = 0;
	list.push_back(this);
}

void TrackFileNTuple::appendRow(float data[], int n)
{
	assert(n == NTrackFileFields);

	G4ThreeVector pos(data[0]*mm,data[1]*mm,data[2]*mm);
	G4ThreeVector momentum(data[3]*MeV,data[4]*MeV,data[5]*MeV);
	G4double time = data[6]*ns;
	int PDGid = (int)data[7];
	int eventID = (int)data[8];
	int trackID = (int)data[9];
	int parentID = (int)data[10];
	double weight = data[11];
	tf->write(pos,time,momentum,PDGid,eventID,trackID,parentID,weight);

	++entries;
}





/**	class FOR009NTuple implements a FOR009.DAT NTuple.
 *
 *	If multiple instances specify different filename-s, the first one wins.
 **/
const int NTraceFileFields=18;
class FOR009NTuple : public BLNTuple {
	G4String title;
	int entries;
	static BLFOR009 *for009file;
	static std::vector<FOR009NTuple*> for009list;
public:
	FOR009NTuple(G4String category, G4String name, G4String fields,
							G4String filename);
	~FOR009NTuple() { }
	virtual void appendRow(float data[], int n);
	virtual bool readRow(float data[], int ndata) { return false; }
	virtual int getNData() { return NTraceFileFields; }
	virtual bool needsReference() { return true; }
	void flush() { }
	void close();
	void doSummary() { /* close provides summary per region */ }
};
BLFOR009 *FOR009NTuple::for009file = 0;
std::vector<FOR009NTuple*> FOR009NTuple::for009list;

class FOR009NTupleHandler : public BLNTupleHandler {
public:
	FOR009NTupleHandler() { 
		BLNTuple::registerHandler("FOR009",this);
		BLNTuple::registerHandler("FOR009.DAT",this);
		BLNTuple::registerHandler("for009",this);
		BLNTuple::registerHandler("for009.dat",this);
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		return new FOR009NTuple(category,name,fields,filename);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields,G4String filename) {
		return 0;
	}
};
FOR009NTupleHandler defaultFOR009NTupleHandler;


FOR009NTuple::FOR009NTuple(G4String category, G4String name,
					G4String fields, G4String filename)
{
	if(for009file == 0) {
		if(filename == "") filename = name+".txt";
		if(filename.find(".txt") == filename.npos)
			filename += ".txt";
		for009file = new BLFOR009(filename,category+"/"+name,"w");
	}
	title = name;
	entries = 0;
	for009list.push_back(this);
}

void FOR009NTuple::appendRow(float data[], int n)
{
	assert(n == NTraceFileFields);

	G4ThreeVector pos(data[0]*mm,data[1]*mm,data[2]*mm);
	G4ThreeVector momentum(data[3]*MeV,data[4]*MeV,data[5]*MeV);
	G4ThreeVector Bfield(data[12]*tesla,data[13]*tesla,data[14]*tesla);
	G4ThreeVector Efield(data[15]*(megavolt/meter),
			data[16]*(megavolt/meter),data[17]*(megavolt/meter));
	G4double time = data[6]*ns;
	int PDGid = (int)data[7];
	int eventID = (int)data[8];
	int trackID = (int)data[9];
	int parentID = (int)data[10];
	double weight = data[11];
	for009file->write(pos,time,momentum,Bfield,Efield,PDGid,eventID,trackID,
							parentID,weight);

	++entries;
}

void FOR009NTuple::close()
{
	if(for009file) for009file->close();
	for009file = 0;
}




/**	class AsciiNTuple implements a generic ASCII NTuple.
 **/
class AsciiNTuple : public BLNTuple {
	FILE *fd;
	G4String title;
	int entries;
	int nfields;
	G4String fields;
public:
	AsciiNTuple(G4String category, G4String name, G4String fields,
							G4String filename);
	~AsciiNTuple() { close(); }
	virtual void appendRow(float data[], int n);
	virtual bool readRow(float data[], int ndata) { return false; }
	virtual int getNData() { return nfields; }
	virtual G4String getFieldNames() { return fields; }
	virtual void annotate(G4String line) 
		{ fprintf(fd,"%s\n",line.c_str()); }
	void flush() { fflush(fd); }
	void close() { if(fd != 0) fclose(fd); fd = 0; }
	void doSummary() {
		printf("NTuple %-15s %8d entries\n",title.c_str(), entries);
	}
};

class AsciiNTupleHandler : public BLNTupleHandler {
public:
	AsciiNTupleHandler() { 
		BLNTuple::registerHandler("Ascii",this);
		BLNTuple::registerHandler("ascii",this);
		BLNTuple::registerHandler("ASCII",this);
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		return new AsciiNTuple(category,name,fields,filename);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields,G4String filename) {
		return 0;
	}
};
AsciiNTupleHandler defaultAsciiNTupleHandler;

AsciiNTuple::AsciiNTuple(G4String category, G4String name,
					G4String _fields, G4String filename)
{
	if(filename == "") filename = name+".txt";
	if(filename.find(".txt") == filename.npos)
		filename += ".txt";
	fd = fopen(filename.c_str(),"w");
	title = name;
	entries = 0;
	fields = _fields;

	fprintf(fd,"# %s\n",name.c_str());
	G4String s(fields);
	nfields = 1;
	for(;;) {
		unsigned p = s.find(":");
		if(p == s.npos) break;
		s.replace(p,1," ");
		++nfields;
	}
	fprintf(fd,"#%s\n",s.c_str());
}

void AsciiNTuple::appendRow(float data[], int n)
{
//printf("AsciiNTuple::appendRow() n=%d\n",n);
	for(int i=0; i<n; ++i)
		fprintf(fd,"%.6g ",data[i]);
	fprintf(fd,"\n");
	++entries;
}

