//	g4bl-histoscope.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/

#ifdef HISTOSCOPE

#include <stdio.h>
#include <vector>
#include <time.h>
#include <assert.h>

#include "BLNTuple.hh"
#include "BLParam.hh"
#include "BLTrackFile.hh"
#include "BLFOR009.hh"
#include "BLCommand.hh"

#include "histoscope.h"

/**	class HistoNTuple implements a HistoScope NTuple.
 **/
class HistoNTuple : public BLNTuple {
	int id;
	int nfields;
	int nEntries;
	int readIndex;
	G4String title;
	static int nextUID;
	friend class HistoNTupleHandler;
public:
	HistoNTuple(int _id, int _nfields, int _nEntries, G4String _title) {
		id=_id; nfields=_nfields; nEntries=_nEntries; readIndex=0;
		title=_title;
	}
	HistoNTuple(G4String category, G4String name, G4String fields);
	~HistoNTuple() { }
	virtual void appendRow(float data[], int n);
	virtual bool readRow(float data[], int ndata);
	virtual int getNData() { return nfields; }
	void flush();
	void close();
	void doSummary();
};
int HistoNTuple::nextUID=0;

class HistoNTupleHandler : public BLNTupleHandler {
public:
	HistoNTupleHandler() {
		BLNTuple::registerHandler("histo",this); 
		BLNTuple::registerHandler("HistoScope",this); 
		BLNTuple::registerHandler("histoscope",this); 
	}
	BLNTuple *create(G4String type, G4String category,
			G4String name, G4String fields, G4String filename) {
		return new HistoNTuple(category,name,fields);
	}
	BLNTuple *read(G4String type, G4String category, G4String name,
					G4String fields, G4String filename);
};
HistoNTupleHandler defaultHistoNTupleHandler;


HistoNTuple::HistoNTuple(G4String category, G4String name, G4String fields)
{
	title = category + "/" + name;
	if(nextUID == 0) {
		nextUID = 1;
		hs_initialize((char *)"g4beamline");
	}

	// split fields into an array
	std::vector<char*> fieldnames;
	char *p = strdup(fields.c_str()); // let it leak
	while(p != 0 && *p != '\0') {
		fieldnames.push_back(p);
		p = strchr(p,':');
		if(p != 0) *p++ = '\0';
	}
	nfields = fieldnames.size();
	nEntries = 0;
	readIndex = 0;

	id = hs_create_ntuple(nextUID++,(char *)name.c_str(),
		(char *)category.c_str(),nfields,&fieldnames[0]);
}

void HistoNTuple::appendRow(float data[], int n)
{
	assert(id>0 && n==nfields);
	hs_fill_ntuple(id,data);
	++nEntries;
}

BLNTuple *HistoNTupleHandler::read(G4String type, G4String category,
			G4String name, G4String fields, G4String filename)
{
	if(HistoNTuple::nextUID == 0) {
		HistoNTuple::nextUID = 1;
		hs_initialize((char *)"g4beamline");
	}

	int uid = atoi(name.c_str());
	int nitems = hs_read_file_items((char *)filename.c_str(),
		(char *)"InputBeam", (char *)category.c_str(), &uid,1);
	if(nitems <= 0) {
		BLCommand::printError("NTuple: cannot open HistoScope file "
			"'%s'",filename.c_str());
		return 0;
	}
	G4String inputCategory("InputBeam");
	if(category != "") {inputCategory += "/"; inputCategory += category;}
	int id = hs_id(uid, (char *)inputCategory.c_str());
	if(id <= 0 || hs_type(id) != HS_NTUPLE) {
		char tmp[256];
		sprintf(tmp,"uid=%d in '%s' of HistoScope file '%s'",
				uid,category.c_str(),filename.c_str());
		G4Exception("HistoNTuple","Cannot find NTuple",FatalException,
									tmp);
	}
	int nEntries = hs_num_entries(id);
	int nfields = hs_num_variables(id);
	return new HistoNTuple(id,nfields,nEntries,inputCategory+"/uid="+name);
}

bool HistoNTuple::readRow(float data[], int ndata)
{
	assert(id>0 && ndata==nfields);
	if(readIndex >= nEntries) return false;
	hs_row_contents(id, readIndex++, data);
	return true;
}

void HistoNTuple::flush()
{
	// flush only once, not evey NTuple, only if some exist
	if(nextUID == 0 || id != 1) return;

	G4String filename = Param.getString("histoFile");
	if(filename == "") return;

	if(filename.find(".hst") == filename.npos)
		filename += ".hst";
	hs_save_file((char *)filename.c_str());
	printf("NTuple wrote HistoScope File '%s'\n",filename.c_str());
}

void HistoNTuple::close()
{
	// close only once, not evey NTuple, only if some exist
	if(nextUID == 0 || id != 1) return;

	flush();
}

void HistoNTuple::doSummary()
{
	if(nextUID == 0) return;

	char title[1024];
	hs_title(id,title);
	printf("NTuple %-15s %8d entries\n",title,hs_num_entries(id));
}

#else
int Histoscope_dummy = 0;
#endif // HISTOSCOPE
