//	BLCMDpillbox2.cc
/*
This source file is part of G4beamline, http://g4beamline.muonsinc.com
Copyright (C) 2003,2004,2005,2006 by Tom Roberts, all rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

http://www.gnu.org/copyleft/gpl.html
*/
//
//      7/18/04 Add model 2 and 3, ky
//     10/10/04 found problem on tuneState
//     10/11/04 Fix rf position helicity to RIGHT HANDED

#include <vector>
#include <stdio.h>

#include "G4VisAttributes.hh"
#include "G4Tubs.hh"
#include "G4LogicalVolume.hh"
#include "G4VPhysicalVolume.hh"
#include "G4PVPlacement.hh"
#include "G4Color.hh"
#include "G4UserLimits.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4SteppingManager.hh"
#include "G4Material.hh"

#include "gsl/gsl_sf_bessel.h"

#include "BLElement.hh"
#include "BLElementField.hh"
#include "BLGlobalField.hh"
#include "BLManager.hh"
#include "BLParam.hh"

/* pitch convention */
#define  LEFT_HANDED_THREAD   (-1)
#define RIGHT_HANDED_THREAD   (+1)

#define MuonsInc_PITCH_CONVENTION    RIGHT_HANDED_THREAD

const double UNDETERMINED = -4.7e21;	// large, negative, unlikely value
inline bool isUndetermined(G4double v) { return v == UNDETERMINED; }

const G4int ITERATION_LIMIT = 100;	// limit on iterations, for both
					// setting timeOffset and tuning

class BLCMDpillbox2; // forward reference

/**	Pillbox2Field represents one placement of a pillbox2.
 *
 *	frequency=0.0 is accepted and yields a pillbox2 with a constant
 *	Ez = maxGradient (useful to verify units are correct).
 *	d.c. cavities cannot be tuned.
 **/
class Pillbox2Field : public BLElementField, public BLManager::SteppingAction {
	BLCMDpillbox2 *pillbox2;
	G4VPhysicalVolume *timingPV;
	BLCoordinateTransform global2local;
	G4RotationMatrix *rotation;
	G4double timeOffset;
        G4double model;
	G4Track saveTrack;
	G4int timeCount;
	G4bool validSaveTrack;
	G4bool settingTiming;
	G4bool isTune;
	friend class BLCMDpillbox2;
public:
	/// constructor. _zmin/max are global coordinates.
	Pillbox2Field(BLCoordinateTransform& _global2local,
			BLCMDpillbox2 *_pillbox2, G4VPhysicalVolume *_timingPV);

	/// addFieldValue() adds the field for this pillbox2 into field[].
	/// point[] is in global coordinates.
	void addFieldValue(const G4double point[4], G4double field[6]) const;

	/// doStep() handles a step of the tune particle. 
	/// Sets timeOffset, and handles tuning (if enabled).
	void UserSteppingAction(const G4Step *step);
};

/**	BLCMDpillbox2 implements a pillbox2.
 *
 *	Each placement of a BLCMDpillbox2 creates a Pillbox2Field that is linked
 *	into BLGlobalField. Each Pillbox2Field sets its timeOffset via the
 *	tune particle (re-stepping through its TimingVol if necessary).
 *
 *	If tuneEnd is nonnull, the Pillbox2Field of that name will automatically
 *	tune the maxGradient of the BLCMDpillbox2 so the energy of the tune
 *	particle at its upstream edge is Efirst * goalFactor + goalOffset.
 *	Here Efirst is the energy of the tune particle at the upstream
 *	edge of the first pillbox2 of this kind. This tuning is done using a
 *	simple, linear, solver algorithm that typically converges to
 *	0.001 MeV in 5 iterations or so.
 **/
class BLCMDpillbox2 : public BLElement {
	enum TuneState {NOTUNE, NEEDTUNE, TUNING, GOODTUNE};
	G4double maxGradient;
	G4String color;
	G4double frequency;
	G4double innerLength;
	G4double pipeThick;
	G4double wallThick;
	G4double irisRadius;
	G4double collarRadialThick;
	G4double collarThick;
	G4double win1Thick;
	G4double win1OuterRadius;
	G4double win2Thick;
	G4String winMat;
	G4double phaseAcc;
	G4double skinDepth;
	G4double timingTolerance;
	G4double maxStep;
	G4String tuneEnd;
	G4double goalFactor;
	G4double goalOffset;
	G4double initialStep;
	G4double tuneTolerance;
	G4String cavityMaterial;
	G4double timeOffset;
        G4double model;
	// non-arguments:
	G4double innerRadius;
	G4double overallLength;
	G4double outerRadius;
	G4double omega;
	G4double rFactor;
	G4double E2Bfactor;
	Pillbox2Field *tuneFirst;
	Pillbox2Field *tuneLast;
	G4double prevE;
	G4double prevGrad;
	G4int tuneCount;
	TuneState tuneState;
	std::vector<Pillbox2Field *> pillbox2Field;
	friend class Pillbox2Field;
public:
	/// Default constructor. Defines the command, etc.
	BLCMDpillbox2();

	/// Destructor.
	virtual ~BLCMDpillbox2() { }

	/// Copy constructor.
	BLCMDpillbox2(const BLCMDpillbox2& r);

	/// clone()
	BLElement *clone() { return new BLCMDpillbox2(*this); }

	/// commandName() returns "pillbox2".
	G4String commandName() { return "pillbox2"; }

	/// command() implements the pillbox2 command.
	int command(BLArgumentVector& argv, BLArgumentMap& namedArgs);

	/// defineNamedArgs() defines the named arguments of the command.
	void defineNamedArgs();

	/// argChanged() does internal computations after some arg changed
	void argChanged();

	/// construct() - construct the pillbox2.
	/// Creates a new Pillbox2Field and adds it to BLGlobalField.
	void construct(G4RotationMatrix *relativeRotation,
			G4ThreeVector relativePosition, 
			G4LogicalVolume* parent,
			G4String parentName,
			G4RotationMatrix *parentRotation,
			G4ThreeVector parentPosition);

	/// getLength() returns the overallLength of the pillbox2
	G4double getLength() { return overallLength; }

	/// getWidth() returns the outer radius of the pillbox2
	G4double getWidth() { return outerRadius; }

	/// getHeight() returns the outer radius of the pillbox2
	G4double getHeight() { return outerRadius; }

	/// isOK() returns true only if all placed pillbox2es of this
	/// type have had their timing set, and if the tuning (if
	/// used) converged.
	G4bool isOK();

	/// generatePoints() from BLElement
	void generatePoints(int npoints, std::vector<G4ThreeVector> &v);

	/// isOutside() from BLElement
	G4bool isOutside(G4ThreeVector &local, G4double tolerance);
};

BLCMDpillbox2 defaultPillbox2;	// default object

// Default constructor - be sure to use the default constructor BLElement()
BLCMDpillbox2::BLCMDpillbox2() : BLElement(), pillbox2Field()
{
	// register the commandName(), and its synopsis and description.
	registerCommand(BLCMDTYPE_ELEMENT);
	setSynopsis("Defines a pillbox2 RF cavity");
	setDescription("A Pillbox2 RF cavity is the basic RF element used to\n"
		"construct a linac. The phaseAcc parameter sets the\n"
		"phase of the tune particle at the center of the\n"
		"cavity, and the timing offset of the cavity is determined\n"
		"from that when the tune particle is tracked through\n"
		"the cavity. Zero degrees is the rising zero-crossing of\n"
		"the Ez field. If tuneEnd is non-null, it must be the name\n"
		"of a later pillbox2 in the beamline; the cavities of this\n"
		"type up to that cavity will be used to tune maxGradient\n"
		"for ALL cavities of this type. The tuning criterion is\n"
		"that the tune particle's momentum when entering the\n"
		"tuneEnd cavity is related to its momentum when it entered\n"
		"the first cavity of this type by goalFactor and goalOffset.\n"
		"Pipe, wall, collar, win1, and win2 can be omitted by setting\n"
		"their thickness to 0.\n"
		"NOTE: using pillbox2 arguments in the place command will\n"
		"prevent any tuning of maxGradient.");

	// provide initial values for fields
	maxGradient = UNDETERMINED; // will be converted to MV/m
	color = "1.0,0.0,0.0";
	frequency = UNDETERMINED;
	innerLength = UNDETERMINED;
	pipeThick = 3.0*mm;
	wallThick = 3.0*mm;
	irisRadius = 11.0*cm;
	collarRadialThick = 5.0*mm;
	collarThick = 2.5*mm;
	win1Thick = 0.200*mm;
	win1OuterRadius = 5.0*cm;
	win2Thick = 0.500*mm;
	winMat = "Be";
	phaseAcc = 40.0*deg;
	skinDepth = 0.002*mm;
	timingTolerance = 0.001*ns;
	maxStep = -1.0;
	tuneEnd = "";
	goalFactor = 1.0;
	goalOffset = 0.0;
	initialStep = 0.2;	// will be converted to MV/m
	tuneTolerance = 0.01;	// will be converted to MV/m
	cavityMaterial = "Vacuum";
	timeOffset = UNDETERMINED;
        model =1.;
	// non-arguments:
	innerRadius = UNDETERMINED;
	overallLength = UNDETERMINED;
	outerRadius = UNDETERMINED;
	omega = UNDETERMINED;
	rFactor = UNDETERMINED;
	E2Bfactor = UNDETERMINED;
	tuneFirst = 0;
	tuneLast = 0;
	prevE = UNDETERMINED;
	prevGrad = UNDETERMINED;
	tuneCount = 0;
	tuneState = NOTUNE;
}

// Copy constructor - be sure to use the copy constructor BLElement(r)
BLCMDpillbox2::BLCMDpillbox2(const BLCMDpillbox2& r) : BLElement(r), pillbox2Field()
{
	// copy fields one at a time (transfers default values from the
	// default object to this new object).
	maxGradient = r.maxGradient;
	color = r.color;
	frequency = r.frequency;
	innerLength = r.innerLength;
	pipeThick = r.pipeThick;
	wallThick = r.wallThick;
	irisRadius = r.irisRadius;
	collarRadialThick = r.collarRadialThick;
	collarThick = r.collarThick;
	win1Thick = r.win1Thick;
	win1OuterRadius = r.win1OuterRadius;
	win2Thick = r.win2Thick;
	winMat = r.winMat;
	phaseAcc = r.phaseAcc;
	skinDepth = r.skinDepth;
	timingTolerance = r.timingTolerance;
	maxStep = r.maxStep;
	tuneEnd = r.tuneEnd;
	goalFactor = goalFactor;
	goalOffset = r.goalOffset;
	initialStep = r.initialStep;
	tuneTolerance = r.tuneTolerance;
	cavityMaterial = r.cavityMaterial;
	timeOffset = r.timeOffset;
	model=r.model;

	// non-arguments:
	overallLength = r.overallLength;
	outerRadius = r.outerRadius;
	omega = r.omega;
	rFactor = r.rFactor;
	E2Bfactor = r.E2Bfactor;
	tuneFirst = r.tuneFirst;
	tuneLast = r.tuneLast;
	prevE = r.prevE;
	prevGrad = r.prevGrad;
	tuneCount = r.tuneCount;
	tuneState = r.tuneState;
	innerRadius = r.innerRadius;
}

int BLCMDpillbox2::command(BLArgumentVector& argv, BLArgumentMap& namedArgs)
{
	if(argv.size() != 1) {
		printError("pillbox2: Invalid command, must have name");
		return -1;
	}

	if(argv[0] == "default") {
		return defaultPillbox2.handleNamedArgs(namedArgs);
	}

	BLCMDpillbox2 *p = new BLCMDpillbox2(defaultPillbox2);
	p->setName(argv[0]);
	int retval = p->handleNamedArgs(namedArgs);

	// check material exists
	if(p->cavityMaterial.size() > 0) getMaterial(p->cavityMaterial,false);

	p->print(argv[0]);

	return retval;
}

void BLCMDpillbox2::defineNamedArgs()
{
	argDouble(maxGradient,"maxGradient","The peak gradient of the cavity (MV/m)",megavolt/meter);
	argString(color,"color","The color of the cavity");
	argDouble(frequency,"frequency","The frequency of the cavity (GHz)",1/ns,"",false);
	argDouble(innerLength,"innerLength","The inside length of the cavity (mm)",mm,"",false);
	argDouble(pipeThick,"pipeThick","The thickness of the pipe wall (mm)",mm,"",false);
	argDouble(wallThick,"wallThick","The thickness of the cavity walls (mm)",mm,"",false);
	argDouble(irisRadius,"irisRadius","The radius of the iris (mm)",mm,"",false);
	argDouble(collarRadialThick,"collarRadialThick","The radial thickness of the collar (mm)",mm,"",false);
	argDouble(collarThick,"collarThick","The thickness of the collar (mm)",mm,"",false);
	argDouble(win1Thick,"win1Thick","The thickness of the inner portion of the\n"
		"windows (mm)",mm,"",false);
	argDouble(win1OuterRadius,"win1OuterRadius","The radius of the inner portion of\n"
		"the windows (mm)",mm,"",false);
	argDouble(win2Thick,"win2Thick","The thickness of the outer portion of the\n"
		"windows (mm)",mm,"",false);
	argString(winMat,"winMat","The material of the windows");
	argDouble(phaseAcc,"phaseAcc","The reference phase of the cavity (degrees)",deg);
	argDouble(skinDepth,"skinDepth","The skin depth (mm)",mm,"",false);
	argDouble(timingTolerance,"timingTolerance","Tolerance for timing tuning (ns)",ns);
	argDouble(maxStep,"maxStep","The maximum stepsize in the element (mm).",mm);
	argString(tuneEnd,"tuneEnd","The name of the last cavity for gradient tuning.",false);
	argDouble(goalFactor,"goalFactor","Tune goal is Pbegin * goalFactor + goalOffset.");
	argDouble(goalOffset,"goalOffset","Tune goal is Pbegin * goalFactor + goalOffset (MeV/c.",MeV);
	argDouble(initialStep,"initialStep","Initial tuning step (MV/m).",megavolt/meter);
	argDouble(tuneTolerance,"tuneTolerance","Tolerance for successful tune (MV/m).",megavolt/meter);
	argString(cavityMaterial,"cavityMaterial","Material of cavity volume (Vacuum).");
	argDouble(model,"model","The model of field calculated (bessel=1, simple icool=2, twister=3)",1,"",false);
	argDouble(timeOffset,"timeOffset","Time offset for cavity (tuned if tune enters) (ns).",ns);
}

void BLCMDpillbox2::argChanged()
{
	// compute non-argument variables
	omega = 2.0 * pi * frequency;
	if(innerRadius == UNDETERMINED) {
		if(frequency > 0.0) {
			// the first zero of J0(r) is at r=2.405
			innerRadius =  2.405 / (omega/c_light);
		} else {
			printError("pillbox: innerRadius is undefined");
		}
	}
	overallLength = innerLength + 2.0 * collarThick;
	outerRadius = innerRadius + pipeThick;
	rFactor = omega/c_light;
	E2Bfactor = 1.0e-9 / c_light;   // MeV/m, MKS units
	if(maxStep < 0.0) maxStep = Param.getDouble("maxStep");
}

void BLCMDpillbox2::construct(G4RotationMatrix *relativeRotation,
			G4ThreeVector relativePosition, 
			G4LogicalVolume* parent,
			G4String parentName,
			G4RotationMatrix *parentRotation,
			G4ThreeVector parentPosition)
{
	G4String thisname = parentName+getName();

	G4Material *cu = getMaterial("Cu");
	G4Material *wm = getMaterial(winMat);
	G4Material *cavityMat = getMaterial(cavityMaterial);
	G4UserLimits *userLimits = new G4UserLimits(maxStep);

	// find enclosing Tubs
	G4double halflen = innerLength/2+wallThick;
	halflen = std::max(halflen,innerLength/2+collarThick);
	halflen = std::max(halflen,innerLength/2+win1Thick);
	halflen = std::max(halflen,innerLength/2+win2Thick);
	G4Tubs *tubs = new G4Tubs(getName(),0.0,innerRadius+pipeThick,
					halflen,0.0,2.0*pi);
	G4LogicalVolume *pillbox2 = new G4LogicalVolume(tubs,cavityMat,
					parentName+getName(),0,0,userLimits);

	G4ThreeVector loc;

	// The outer pipe
	G4LogicalVolume *pipe=0;
	if(pipeThick > 0.001*mm) {
		tubs = new G4Tubs(getName()+"pipe",innerRadius,
					innerRadius+pipeThick,
					innerLength/2+wallThick,0.0,2.0*pi);
		pipe = new G4LogicalVolume(tubs,cu,
					getName()+"pipeLog",0,0,userLimits);
		new G4PVPlacement(0,loc,pipe,parentName+getName()+"Pipe",
					pillbox2,false,0);
	}

	// the wall (us and ds) - radially: collarOuterRadius to innerRadius
	G4LogicalVolume *wall=0;
	if(wallThick > 0.001*mm) {
		tubs = new G4Tubs(getName()+"wall",irisRadius+collarRadialThick,
					innerRadius,wallThick/2,0.0,2.0*pi);
		wall = new G4LogicalVolume(tubs,cu,
					getName()+"wallLog",0,0,userLimits);
		G4double wallCenterZ = innerLength/2.0 + wallThick/2.0;
		loc.setZ(-wallCenterZ);
		new G4PVPlacement(0,loc,wall,parentName+getName()+"UsWall",
					pillbox2,false,0);
		loc.setZ(+wallCenterZ);
		new G4PVPlacement(0,loc,wall,parentName+getName()+"DsWall",
					pillbox2,false,0);
	}
	
	// the collar (us and ds) - includes wall thickness (i.e. the inner
	// edge of the collar is the same as the inner edge of the wall)
	G4LogicalVolume *collar=0;
	if(collarThick > 0.001*mm) {
		tubs = new G4Tubs(getName()+"collar",irisRadius,
					irisRadius+collarRadialThick,
					collarThick/2,0.0,2.0*pi);
		collar = new G4LogicalVolume(tubs,cu,
					getName()+"collarLog",0,0,userLimits);
		G4double collarCenterZ = innerLength/2.0 + collarThick/2.0;
		loc.setZ(-collarCenterZ);
		new G4PVPlacement(0,loc,collar,parentName+getName()+"UsCollar",
					pillbox2,false,0);
		loc.setZ(+collarCenterZ);
		new G4PVPlacement(0,loc,collar,parentName+getName()+"DsCollar",
					pillbox2,false,0);
	}
	
	// win1 (the inner window)
	G4LogicalVolume *win1=0;
	if(win1Thick > 0.001*mm) {
		tubs = new G4Tubs(getName()+"win1",0.0,win1OuterRadius,
					win1Thick/2,0.0,2.0*pi);
		win1 = new G4LogicalVolume(tubs,wm,
					getName()+"win1Log",0,0,userLimits);
		G4double win1CenterZ = innerLength/2.0 + win1Thick/2.0;
		loc.setZ(-win1CenterZ);
		new G4PVPlacement(0,loc,win1,parentName+getName()+"UsWin1",
					pillbox2,false,0);
		loc.setZ(+win1CenterZ);
		new G4PVPlacement(0,loc,win1,parentName+getName()+"DsWin1",
					pillbox2,false,0);
	}

	// win2 (the outer window)
	G4LogicalVolume *win2=0;
	if(win2Thick > 0.001*mm) {
		tubs = new G4Tubs(getName()+"win2",win1OuterRadius,irisRadius,
					win2Thick/2,0.0,2.0*pi);
		win2 = new G4LogicalVolume(tubs,wm,
					getName()+"win2Log",0,0,userLimits);
		G4double win2CenterZ = innerLength/2.0 + win2Thick/2.0;
		loc.setZ(-win2CenterZ);
		new G4PVPlacement(0,loc,win2,parentName+getName()+"UsWin2",
					pillbox2,false,0);
		loc.setZ(+win2CenterZ);
		new G4PVPlacement(0,loc,win2,parentName+getName()+"DsWin2",
					pillbox2,false,0);
	}

	// half of the interior
	tubs = new G4Tubs(getName()+"half",0.0,innerRadius,
					innerLength/4,0.0,2.0*pi);
	G4LogicalVolume *half = new G4LogicalVolume(tubs,cavityMat,getName()+"halfLog",
	 				0,0,userLimits);
	G4double halfCenterZ = innerLength/4.0;
	loc.setZ(+halfCenterZ);
	new G4PVPlacement(0,loc,half,parentName+getName()+"DsHalf",
							pillbox2,false,0);
	loc.setZ(-halfCenterZ);
	G4VPhysicalVolume *timingPV = new G4PVPlacement(0,loc,half,
				parentName+getName()+"TimingVol",pillbox2,false,0);

#ifdef G4VIS_USE
	G4VisAttributes *c1 = new G4VisAttributes(true,G4Color(0.5,0.0,0.0));
	G4VisAttributes *c2 = new G4VisAttributes(true,G4Color(0.7,0.0,0.0));
	G4VisAttributes *c3 = new G4VisAttributes(true,G4Color(1.0,0.0,0.0));
	const G4VisAttributes *invisible = &G4VisAttributes::GetInvisible();
	if(pipe) pipe->SetVisAttributes(c1);
	if(wall) wall->SetVisAttributes(c1);
	if(collar) collar->SetVisAttributes(c1);
	if(win1) win1->SetVisAttributes(c3);
	if(win2) win2->SetVisAttributes(c2);
	half->SetVisAttributes(invisible);
	pillbox2->SetVisAttributes(invisible);
#endif

	// geant4 rotation convention is backwards from g4beamline
	G4RotationMatrix *g4rot = 0;
	if(relativeRotation)
		g4rot = new G4RotationMatrix(relativeRotation->inverse());

	// place the pillbox2 into the parent
	new G4PVPlacement(g4rot,relativePosition,pillbox2,
				parentName+getName(),parent,false,0);

	// get globalRotation and globalPosition
	G4RotationMatrix *globalRotation = 0;
	if(relativeRotation && parentRotation) {
		globalRotation = 
		    new G4RotationMatrix(*parentRotation * *relativeRotation);
	} else if(relativeRotation) {
		globalRotation = relativeRotation;
	} else if(parentRotation) {
		globalRotation = parentRotation;
	}
	G4ThreeVector globalPosition(relativePosition + parentPosition);
	if(parentRotation)
		globalPosition = *parentRotation * relativePosition +
				parentPosition;

	BLCoordinateTransform global2local(globalRotation,globalPosition);

	// add this pillbox2 field to the GlobalField, and to BLManager
	G4double dz = innerLength/2.0 + wallThick;
	// if rotated, make an overestimate of the field occupancy along z
	if(global2local.isRotated())
		dz += innerRadius;
	Pillbox2Field *pf = new Pillbox2Field(global2local,this,timingPV);
	BLGlobalField::getObject()->addElementField(pf);
	pillbox2Field.push_back(pf);
	BLManager::getObject()->registerTuneParticleStep(timingPV,pf);

	// handle tuning
	if(tuneEnd != "") {
		// upstream-most BLCMDpillbox2
		if(tuneFirst == 0) {
			tuneFirst = pf;
		}
		// downstream-most tuneEnd
		if(thisname == tuneEnd) {
			if(tuneLast)
				tuneLast->isTune = false;
			tuneLast = pf;
			tuneLast->isTune = true;
		}
		tuneState = NEEDTUNE;
	}

	printf("BLCMDpillbox2::construct %s parent=%s relZ=%.1f globZ=%.1f\n"
			"\tzmin=%.1f zmax=%.1f\n",
		thisname.c_str(),parentName.c_str(),relativePosition[2],
		global2local.getPosition()[2], global2local.getPosition()[2]-dz,
		global2local.getPosition()[2]+dz);
}

G4bool BLCMDpillbox2::isOK()
{
	G4bool retval = true;

	// verify all placed pillbox2Field-s have set their timeOffset 
	std::vector<Pillbox2Field *>::iterator i;
	for(i=pillbox2Field.begin(); i!=pillbox2Field.end(); ++i) {
		if(isUndetermined((*i)->timeOffset) || (*i)->settingTiming) {
			printf("*** BLCMDpillbox2 %s has not determined timing\n",
				getName().c_str());
			retval = false;
		}
	}

	if(tuneState != NOTUNE && tuneState != GOODTUNE) {
		printf("*** BLCMDpillbox2 %s Tuning has not convereged\n",
				getName().c_str());
		retval = false;
	}

	return retval;
}

Pillbox2Field::Pillbox2Field(BLCoordinateTransform& _global2local,
			BLCMDpillbox2 *_pillbox2,
			G4VPhysicalVolume *_timingPV) :
			BLElementField(), BLManager::SteppingAction()
{
	pillbox2 = _pillbox2;
	timingPV = _timingPV;
	global2local = _global2local;
	rotation = 0;
	timeOffset = pillbox2->timeOffset;
	model=pillbox2->model;
	timeCount = 0;
	validSaveTrack = false;
	settingTiming = false;
	isTune = false;

	if(global2local.isRotated()) {
		rotation = new G4RotationMatrix(global2local.getRotation());
		rotation->invert();
	}

	if(pillbox2->frequency == 0.0) {
		timeOffset = 0.0;
		settingTiming = false;
	}

	// set global bounding box
	G4double local[4], global[4];
	local[3] = 0.0;
	G4double dz = pillbox2->innerLength/2.0 + pillbox2->wallThick;
	for(int i=0; i<2; ++i) {
		local[0] = (i==0 ? -1.0 : 1.0) * pillbox2->innerRadius;
		for(int j=0; j<2; ++j) {
			local[1] = (j==0 ? -1.0 : 1.0) * pillbox2->innerRadius;
			for(int k=0; k<2; ++k) {
				local[2] = (k==0 ? -1.0 : 1.0) * dz;
				global2local.getGlobal(local,global);
				setGlobalPoint(global);
			}
		}
	}
}

void Pillbox2Field::addFieldValue(const G4double point[4], G4double field[6]) const
{
	G4double local[4];
	global2local.getLocal(local,point);
	G4double x = local[0];
	G4double y = local[1];
	G4double z = local[2];
	G4double r = sqrt(x*x + y*y);
	G4double dz = pillbox2->innerLength/2.0 + pillbox2->wallThick;
	G4double j0=1.0;
	G4double j1=1.0;
	G4double bphi=0.0;

	if(z < -dz || z > dz || r > pillbox2->innerRadius) return;

	// frequency == 0.0 means a d.c. electric field
	if(pillbox2->frequency == 0.0 && int(model)==1 ) {
		// zmin and zmax include the walls, so re-test z position
		if(z >= -pillbox2->innerLength/2.0 && 
		   z <=  pillbox2->innerLength/2.0)
			field[5] += pillbox2->maxGradient;
		return;
	}

	// compute normal RF pillbox2 field
	double tArg = pillbox2->omega * (point[3] - timeOffset);

	if( int(model)==1 ){
	  j0   = gsl_sf_bessel_J0(pillbox2->rFactor*r);
	  j1   = gsl_sf_bessel_J1(pillbox2->rFactor*r);
	  bphi = -pillbox2->maxGradient*pillbox2->E2Bfactor*j1*sin(tArg);
	}
	else if ( int(model) == 2 ){}

	double ez = pillbox2->maxGradient * j0 * cos(tArg);
	// handle skin depth if in wall or window
	double f = 1.0;
	if(z > pillbox2->innerLength/2.0)
		f=exp(-(z-pillbox2->innerLength/2.0)/pillbox2->skinDepth);
	else if(z < -pillbox2->innerLength/2.0)
		f=exp(-(-z-pillbox2->innerLength/2.0)/pillbox2->skinDepth);
	double phi = atan2(point[1],point[0]);

	if ( int(model) == 3 ){
	  ez=pillbox2->maxGradient;
	  G4ThreeVector E(ez*cos(phi)*f,ez*sin(phi)*f,ez*f);
	  field[3] += E[0];
	  field[4] += E[1];
	  field[5] += E[2];
	  return;
	}

	if(rotation) {
		G4ThreeVector B(-bphi*sin(phi)*f, bphi*cos(phi)*f, 0.0);
		G4ThreeVector E(0.0, 0.0, ez*f);
		B = *rotation * B;
		E = *rotation * E;
		field[0] += B[0];
		field[1] += B[1];
		field[2] += B[2];
		field[3] += E[0];
		field[4] += E[1];
		field[5] += E[2];
	} 
	else {
		field[0] += -bphi * sin(phi) * f;
		field[1] += bphi * cos(phi) * f;
		field[5] += ez * f;
	}
}

// called ONLY in tune particle mode, for this pillbox2's timingVol
void  Pillbox2Field::UserSteppingAction(const G4Step *step)

{
    // do nothing if d.c.
    if(pillbox2->frequency <= 0.0) return;

    G4Track *track = step->GetTrack();
    G4StepPoint *prePoint = step->GetPreStepPoint();
    if(!prePoint) return;
    G4VPhysicalVolume *prePV = prePoint->GetPhysicalVolume();
    G4StepPoint *postPoint = step->GetPostStepPoint();
    if(!postPoint) return;
    G4VPhysicalVolume *postPV = postPoint->GetPhysicalVolume();
    G4SteppingManager *steppingMgr =
			BLManager::getObject()->getSteppingManager();

    if(prePV == postPV) return;     // neither entering nor leaving

    if(postPV == timingPV) {	    // entering timingPV
        if(isTune) {
            // handle tuning maxGradient
            Pillbox2Field *first = pillbox2->tuneFirst;
            if(!first) {
                printf("BLCMDpillbox2::tuneParticleStep NO FIRST PILLBOX -"
				" aborting\n");
                fflush(stdout);
                ::abort();
            }
            pillbox2->tuneState = BLCMDpillbox2::TUNING;
            if(pillbox2->prevE > 0.0) {
                // update maxGradient
                G4double firstE = first->saveTrack.GetKineticEnergy();
                G4double desiredE = firstE * pillbox2->goalFactor +
                            pillbox2->goalOffset;
                G4double thisE = track->GetKineticEnergy();
                G4double prevE = pillbox2->prevE;
                G4double thisGrad = pillbox2->maxGradient;
                G4double prevGrad = pillbox2->prevGrad;
                G4double slope = (thisE - prevE) / (thisGrad - prevGrad);
                if(fabs(slope) < 1.0e-66) {
                    printf("BLCMDpillbox2::tuneParticleStep CANNOT TUNE"
		    	" Pillbox2 %s slope=%g\n",
			   pillbox2->getName().c_str(), slope);
		    printf("BLCMDpillbox2:: thisE=%f prevE=%f\n",
			   thisE, prevE);
		    printf("BLCMDpillbox2:: thisGrad=%f prevGrad=%f\n",
			   thisGrad, prevGrad);
		    extern void g4bl_exit(int);
		    g4bl_exit(0);
                }
                G4double delta = (desiredE - thisE) / slope;
                if(delta < -10.0*fabs(pillbox2->initialStep))
			delta = -10.0*fabs(pillbox2->initialStep);
                if(delta > 10.0*fabs(pillbox2->initialStep))
			delta = 10.0*fabs(pillbox2->initialStep);
                pillbox2->prevE = thisE;
                pillbox2->prevGrad = thisGrad;
                pillbox2->maxGradient += delta;
		printf("BLCMDpillbox2 '%s' Tuning  maxGradient=%.3f MV/m\n",
				pillbox2->getName().c_str(),
				pillbox2->maxGradient/(megavolt/meter));
                if(fabs(delta) > pillbox2->tuneTolerance) {
                    if(++pillbox2->tuneCount > ITERATION_LIMIT) {
                      printf("BLCMDpillbox2::tuneParticleStep ITERATION LIMIT"
		    	    " BLCMDpillbox2 %s\n",pillbox2->getName().c_str());
                      ::_exit(1);
                    }
                    if(!first->validSaveTrack) {
                        fflush(stdout);
                        fprintf(stderr,"BLCMDpillbox2::tuneParticleStep "
                            "ERROR - attempt to restore invalid step\n");
                        fflush(stderr);
                        ::abort();
                    }
		    // jump back to the upstream edge of the first cavity
                    steppingMgr->GetSecondary()->
                        push_back(new G4Track(first->saveTrack));
                    track->SetTrackStatus(fStopAndKill);
                    return;
                }
		// done tuning!
                pillbox2->tuneState = BLCMDpillbox2::GOODTUNE;
            } else {
                // first track - force a second
                pillbox2->prevGrad = pillbox2->maxGradient;
                pillbox2->prevE = track->GetKineticEnergy();
                pillbox2->maxGradient += pillbox2->initialStep;
                if(!first->validSaveTrack) {
                    fflush(stdout);
                    fprintf(stderr,"BLCMDpillbox2::tuneParticleStep "
                        "ERROR - attempt to restore invalid step\n");
                    fflush(stderr);
                    ::abort();
                }
                // restore track from first Pillbox2Field
                steppingMgr->GetSecondary()->
                    push_back(new G4Track(first->saveTrack));
                track->SetTrackStatus(fStopAndKill);
                return;
            }
        }
        // save track for setting timeOffset (and tuning if this is first)
        saveTrack.CopyTrackInfo(*step->GetTrack());
	saveTrack.SetUserInformation(0);
        validSaveTrack = true;
        // estimate timingOffset by neglecting acceleration
        if(!settingTiming) {
            settingTiming = true;
            double v = saveTrack.GetVelocity();
            double dist = pillbox2->innerLength/2.0;
            double arrival = saveTrack.GetGlobalTime() + dist/v;
            // pi/2 because E is a cos() and we want phaseAcc=0
            // to be the rising zero-crossing
            timeOffset = arrival -
                (pillbox2->phaseAcc-pi/2.0)/pillbox2->omega;
        }
    } else if(prePV == timingPV) {	    // leaving timingPV
        if(++timeCount > ITERATION_LIMIT) {
            printf("BLCMDpillbox2::tuneParticleStep CANNOT TIME"
	    	" BLCMDpillbox2 %s timeCount=%d\n",
		   pillbox2->getName().c_str(),timeCount);
            ::_exit(1);
        }
        // check timing offset
        double t = track->GetGlobalTime();
        double phase = pillbox2->omega * (t-timeOffset) + pi/2.0;
        double dt = (phase-pillbox2->phaseAcc)/pillbox2->omega;
        timeOffset = timeOffset + dt;
        if(fabs(dt) <= pillbox2->timingTolerance) {
            // Success!
            settingTiming = false;
            return;
        }
        // restore saved data (i.e. jump back to when the track 
        // enteered TimingVol)
        if(!validSaveTrack) {
            fflush(stdout);
            fprintf(stderr,"BLCMDpillbox2::tuneParticleStep ERROR"
                " - attempt to restore invalid step\n");
            fflush(stderr);
            ::abort();
        }
        steppingMgr->GetSecondary()->push_back(new G4Track(saveTrack));
        track->SetTrackStatus(fStopAndKill);
    }
}

void BLCMDpillbox2::generatePoints(int npoints, std::vector<G4ThreeVector> &v)
{
	generateTubs(npoints, 0.0, outerRadius, 0.0, 360.0*deg,
			overallLength, v);
}

G4bool BLCMDpillbox2::isOutside(G4ThreeVector &local, G4double tolerance)
{
	G4double r = sqrt(local[0]*local[0]+local[1]*local[1]);
	return r > outerRadius-tolerance ||
		fabs(local[2]) > overallLength/2.0-tolerance;
}
